# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""Small module with helper functions to analyze and manipulate Python abstract syntax tree"""
import k1lib, math, numpy as np, random, base64, json, time, ast; from typing import List, Iterator
from collections import defaultdict, deque; import k1lib.cli as cli
__all__ = ["walk", "pretty", "py2js", "asyncGuard"]
class JSTranspileError(Exception): pass                                          # JSTranspileError
def _walk(m, parent, accessor="", depth=0): # yields (elem, parent, parent accessor, depth). This function don't need to recurse leaf nodes, which is a pretty good time saver # _walk
    yield m, parent, accessor, depth                                             # _walk
    if isinstance(m, (ast.Module, ast.ClassDef, ast.FunctionDef)):               # _walk
        for expr in m.body: yield from _walk(expr, m, f".body[{len(m.body)}]", depth+1) # _walk
    elif isinstance(m, ast.Expr):                                                # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
    elif isinstance(m, ast.Lambda):                                              # _walk
        yield from _walk(m.body, m, ".body", depth+1)                            # _walk
    elif isinstance(m, ast.BinOp):                                               # _walk
        yield from _walk(m.left, m, ".left", depth+1)                            # _walk
        yield from _walk(m.op, m, ".op", depth+1)                                # _walk
        yield from _walk(m.right, m, ".right", depth+1)                          # _walk
    elif isinstance(m, ast.Assign):                                              # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
        for e in m.targets: yield from _walk(e, m, f".targets[{len(m.targets)}]", depth+1) # _walk
    elif isinstance(m, ast.If):                                                  # _walk
        yield from _walk(m.test, m, ".test", depth+1)                            # _walk
        for e in m.body: yield from _walk(e, m, f".body[{len(m.body)}]", depth+1) # _walk
        for e in m.orelse: yield from _walk(e, m, f".orelse[{len(m.orelse)}]", depth+1) # _walk
    elif isinstance(m, ast.Call):                                                # _walk
        yield from _walk(m.func, m, ".func", depth+1)                            # _walk
        for e in m.args: yield from _walk(e, m, f".args[{len(m.args)}]", depth+1) # _walk
        for e in m.keywords: yield from _walk(e, m, f".keywords[{len(m.keywords)}]", depth+1) # _walk
    elif isinstance(m, ast.Attribute):                                           # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
    elif isinstance(m, ast.UnaryOp):                                             # _walk
        yield from _walk(m.op, m, ".op", depth+1)                                # _walk
        yield from _walk(m.operand, m, ".operand", depth+1)                      # _walk
    elif isinstance(m, ast.Subscript):                                           # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
        yield from _walk(m.slice, m, ".slice", depth+1)                          # _walk
        yield from _walk(m.ctx, m, ".ctx", depth+1)                              # _walk
    # elif isinstance(m, ast.Store):                                             # _walk
    #     yield from _walk(m.parent, m, ".parent", depth+1)                      # _walk
    elif isinstance(m, ast.Compare):                                             # _walk
        yield from _walk(m.left, m, ".left", depth+1)                            # _walk
        for e in m.ops: yield from _walk(e, m, f".ops[{len(m.ops)}]", depth+1)   # _walk
        for e in m.comparators: yield from _walk(e, m, f".comparators[{len(m.comparators)}]", depth+1) # _walk
    elif isinstance(m, ast.List):                                                # _walk
        for e in m.elts: yield from _walk(e, m, f".elts[{len(m.elts)}]", depth+1) # _walk
    elif isinstance(m, ast.Tuple):                                               # _walk
        for e in m.elts: yield from _walk(e, m, f".elts[{len(m.elts)}]", depth+1) # _walk
    elif isinstance(m, ast.Set):                                                 # _walk
        for e in m.elts: yield from _walk(e, m, f".elts[{len(m.elts)}]", depth+1) # _walk
    elif isinstance(m, ast.comprehension):                                       # _walk
        yield from _walk(m.is_async, m, ".is_async", depth+1)                    # _walk
        yield from _walk(m.iter, m, ".iter", depth+1)                            # _walk
        yield from _walk(m.target, m, ".target", depth+1)                        # _walk
        for e in m.ifs: yield from _walk(e, m, f".ifs[{len(m.ifs)}]", depth+1)   # _walk
    elif isinstance(m, (ast.ListComp, ast.GeneratorExp, ast.SetComp)):           # _walk
        yield from _walk(m.elt, m, ".elt", depth+1)                              # _walk
        for e in m.generators: yield from _walk(e, m, f".generators[{len(m.generators)}]", depth+1) # _walk
    elif isinstance(m, ast.DictComp):                                            # _walk
        yield from _walk(m.key, m, ".key", depth+1)                              # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
        for e in m.generators: yield from _walk(e, m, f".generators[{len(m.generators)}]", depth+1) # _walk
    elif isinstance(m, ast.Starred):                                             # _walk
        yield from _walk(m.ctx, m, ".ctx", depth+1)                              # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
    elif isinstance(m, ast.keyword):                                             # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
    elif isinstance(m, ast.Dict):                                                # _walk
        for e in m.keys: yield from _walk(e, m, f".keys[{len(m.keys)}]", depth+1) # _walk
        for e in m.values: yield from _walk(e, m, f".values[{len(m.values)}]", depth+1) # _walk
    elif isinstance(m, ast.IfExp):                                               # _walk
        yield from _walk(m.body, m, ".body", depth+1)                            # _walk
        yield from _walk(m.test, m, ".test", depth+1)                            # _walk
        yield from _walk(m.orelse, m, ".orelse", depth+1)                        # _walk
    elif isinstance(m, ast.Slice):                                               # _walk
        yield from _walk(m.lower, m, ".lower", depth+1)                          # _walk
        yield from _walk(m.upper, m, ".upper", depth+1)                          # _walk
        yield from _walk(m.step, m, ".step", depth+1)                            # _walk
    elif isinstance(m, ast.FormattedValue):                                      # _walk
        yield from _walk(m.value, m, ".value", depth+1)                          # _walk
    elif isinstance(m, ast.JoinedStr):                                           # _walk
        for e in m.values: yield from _walk(e, m, f".values[{len(m.values)}]", depth+1) # _walk
class walk(cli.BaseCli):                                                         # walk
    def __init__(self):                                                          # walk
        """Walks the AST tree, returning (elem, parent, parent accessor, depth) tuple.
Example::

    "abc.py" | cat() | join("\n") | kast.walk() # returns generator of tuples
"""                                                                              # walk
        pass                                                                     # walk
    def __ror__(self, m:str):                                                    # walk
        if isinstance(m, str): m = ast.parse(m)                                  # walk
        return _walk(m, None, "", 0)                                             # walk
class pretty(cli.BaseCli):                                                       # pretty
    def __init__(self):                                                          # pretty
        """Returns list[str] containing the prettified tree structure of the AST. Mainly
for asthetic, not for parsing by other tools downstream. Example::

    "x+3" | kast.pretty()
    "lambda x: x+3" | kast.pretty()
"""                                                                              # pretty
        pass                                                                     # pretty
    def __ror__(self, it:str):                                                   # pretty
        m = it; res = []                                                         # pretty
        if isinstance(m, str): m = ast.parse(m)                                  # pretty
        for m, parent, accessor, depth in _walk(m, None, "", 0):                 # pretty
            pad = "  "*depth; res.append(f"{pad}{accessor} {m}")                 # pretty
            if isinstance(m, ast.arguments): res[-1] += f" {[a.arg for a in m.args]}" # pretty
            elif isinstance(m, ast.Import): res[-1] += f" .names[{len(m.names)}].name={', '.join([e.name for e in m.names])}" # pretty
            elif isinstance(m, ast.ImportFrom): res[-1] += f" .module={repr(m.module)} .names[{len(m.names)}].name={', '.join([e.name for e in m.names])}" # pretty
            elif isinstance(m, ast.BinOp): res[-1] += f" .op={m.op}"             # pretty
            elif isinstance(m, (ast.ClassDef, ast.FunctionDef)): res[-1] += f" .name={repr(m.name)}" # pretty
            elif isinstance(m, ast.Name): res[-1] += f" .id={repr(m.id)}"        # pretty
            elif isinstance(m, ast.Constant): res[-1] += f" .value={repr(m.value)}" # pretty
            elif isinstance(m, ast.Attribute): res[-1] += f" .attr={repr(m.attr)}" # pretty
            elif isinstance(m, ast.keyword): res[-1] += f" .arg={repr(m.arg)}"   # pretty
        return res                                                               # pretty
jstCallDict = {}                                                                 # pretty
def _jstGuardNArgs(name, ns, f): # common guard across all short and simple functions # _jstGuardNArgs
    ns = set(ns)                                                                 # _jstGuardNArgs
    if len(ns) == 0: jstCallDict[name] = f                                       # _jstGuardNArgs
    else:                                                                        # _jstGuardNArgs
        def inner(m, lcs):                                                       # _jstGuardNArgs
            if len(m.args) not in ns: raise JSTranspileError(f"Function `{name}` only support these number of arguments: {ns}. You have {len(m.args)} arguments total!") # _jstGuardNArgs
            return f(m, lcs)                                                     # _jstGuardNArgs
        jstCallDict[name] = inner                                                # _jstGuardNArgs
_jstGuardNArgs("abs", [1], lambda m, lcs: f"Math.abs({_py2js(m.args[0], lcs)})") # _jstGuardNArgs
_jstGuardNArgs("all", [1], lambda m, lcs: f"{_py2js(m.args[0], lcs)}.every((x) => x)") # _jstGuardNArgs
_jstGuardNArgs("any", [1], lambda m, lcs: f"{_py2js(m.args[0], lcs)}.some((x) => x)") # _jstGuardNArgs
def _jst_divmod(m, lcs):                                                         # _jst_divmod
    if len(m.args) != 2: raise JSTranspileError(f"Only support divmod() with 2 arguments!") # _jst_divmod
    a = _py2js(m.args[0], lcs); b = _py2js(m.args[1], lcs)                       # _jst_divmod
    return f"[Math.floor(a/b), a-b*Math.floor(a/b)]"                             # _jst_divmod
jstCallDict["divmod"] = _jst_divmod                                              # _jst_divmod
def _jst_enumerate(m, lcs):                                                      # _jst_enumerate
    if len(m.args) != 1: raise JSTranspileError("Currently don't support enumerate() with the start parameter") # _jst_enumerate
    return f"{_py2js(m.args[0], lcs)}.map((x, i) => [x, i])"                     # _jst_enumerate
jstCallDict["enumerate"] = _jst_enumerate                                        # _jst_enumerate
_jstGuardNArgs("float", [1], lambda m, lcs: f"parseFloat({_py2js(m.args[0], lcs)})") # _jst_enumerate
def _jst_getattr(m, lcs):                                                        # _jst_getattr
    a = _py2js(m.args[0], lcs); b = _py2js(m.args[1], lcs)                       # _jst_getattr
    if len(m.args) == 2: return f"{a}[{b}]"                                      # _jst_getattr
    else: return f"{a}[{b}] ?? {_py2js(m.args[2], lcs)}"                         # _jst_getattr
jstCallDict["getattr"] = _jst_getattr                                            # _jst_getattr
_jstGuardNArgs("hasattr", [2], lambda m, lcs: f"{_py2js(m.args[0], lcs)}[{_py2js(m.args[1], lcs)}] !== undefined") # _jst_getattr
_jstGuardNArgs("int", [1], lambda m, lcs: f"parseInt({_py2js(m.args[0], lcs)})") # _jst_getattr
_jstGuardNArgs("iter", [1], lambda m, lcs: f"Array.from({_py2js(m.args[0], lcs)})") # _jst_getattr
_jstGuardNArgs("len", [1], lambda m, lcs: f"{_py2js(m.args[0], lcs)}.length")    # _jst_getattr
_jstGuardNArgs("list", [1], lambda m, lcs: f"Array.from({_py2js(m.args[0], lcs)})") # _jst_getattr
def _jst_max(m, lcs):                                                            # _jst_max
    if len(m.args) == 0: raise JSTranspileError("Only support max() with 1 or more parameters") # _jst_max
    if len(m.args) == 1: return f"Math.max(...({_py2js(m.args[0], lcs)}))"       # _jst_max
    return f"Math.max({', '.join([_py2js(e, lcs) for e in m.args])})"            # _jst_max
jstCallDict["max"] = _jst_max                                                    # _jst_max
def _jst_min(m, lcs):                                                            # _jst_min
    if len(m.args) == 0: raise JSTranspileError("Only support min() with 1 or more parameters") # _jst_min
    if len(m.args) == 1: return f"Math.min(...({_py2js(m.args[0], lcs)}))"       # _jst_min
    return f"Math.min({', '.join([_py2js(e, lcs) for e in m.args])})"            # _jst_min
jstCallDict["min"] = _jst_min                                                    # _jst_min
_jstGuardNArgs("pow", [2], lambda m, lcs: f"Math.pow({_py2js(m.args[0], lcs)}, {_py2js(m.args[1], lcs)})") # _jst_min
_jstGuardNArgs("print", [], lambda m, lcs: f"console.log({', '.join([_py2js(e, lcs) for e in m.args])})") # _jst_min
def _jst_range(m, lcs):                                                          # _jst_range
    n = len(m.args)                                                              # _jst_range
    if n == 1: start = None; stop = m.args[0]                                    # _jst_range
    elif n == 2: start, stop = m.args                                            # _jst_range
    else: raise JSTranspileError("Only support range() without the step variable") # _jst_range
    ans = f"[...Array({_py2js(stop, lcs)}).keys()]"                              # _jst_range
    if start is not None: ans = f"{ans}.slice({_py2js(start, lcs)})"             # _jst_range
    return ans                                                                   # _jst_range
jstCallDict["range"] = _jst_range                                                # _jst_range
_jstGuardNArgs("reversed", [1], lambda m, lcs: f"({_py2js(m.args[0], lcs)}).toReversed()") # _jst_range
def _jst_round(m, lcs):                                                          # _jst_round
    if len(m.args) > 2 or len(m.args) <= 0: raise JSTranspileError("Only support round() with 1 or 2 arguments!") # _jst_round
    if len(m.args) == 1: return f"Math.round({_py2js(m.args[0], lcs)})"          # _jst_round
    else: b = _py2js(m.args[1], lcs); return f"Math.round(({_py2js(m.args[0], lcs)})*Math.pow(10, {b})+Number.EPSILON)/Math.pow(10, {b})" # _jst_round
jstCallDict["round"] = _jst_round                                                # _jst_round
_jstGuardNArgs("set", [1], lambda m, lcs: f"new Set({_py2js(m.args[0], lcs)})")  # _jst_round
_jstGuardNArgs("setattr", [3], lambda m, lcs: f"({_py2js(m.args[0], lcs)}[{_py2js(m.args[1], lcs)}] = {_py2js(m.args[2], lcs)})") # _jst_round
# sorted, ignored cause user can just use sort() cli instead! Python's sort() has weird args # _jst_round
_jstGuardNArgs("str", [1], lambda m, lcs: f"({_py2js(m.args[0], lcs)}).toString()") # _jst_round
_jstGuardNArgs("sum", [1], lambda m, lcs: f"({_py2js(m.args[0], lcs)}).reduce((partialSum, a) => partialSum + a, 0)") # _jst_round
_jstGuardNArgs("tuple", [1], lambda m, lcs: f"Array.from({_py2js(m.args[0], lcs)})") # _jst_round
_jstGuardNArgs("type", [1], lambda m, lcs: f"typeof({_py2js(m.args[0], lcs)})")  # _jst_round
_jstGuardNArgs("zip", [], lambda m, lcs: f"[{', '.join([_py2js(e, lcs) for e in m.args])}].transpose()") # _jst_round
def flattenAttribute(x) -> str:                                                  # flattenAttribute
    if isinstance(x, ast.Name): return x.id                                      # flattenAttribute
    elif isinstance(x, ast.Attribute): return f"{flattenAttribute(x.value)}.{x.attr}" # flattenAttribute
    else: raise JSTranspileError("kast.flattenAttribute() only supports attributes that only contains ast.Attribute and ast.Name") # flattenAttribute
def pyLambParse(s:str) -> str: # parses code, get rid of "lambda: " in front if exists # pyLambParse
    # s might contain stringified                                                # pyLambParse
    m = ast.parse(s)                                                             # pyLambParse
    if len(m.body) == 0: raise JSTranspileError(f"No expression inside passed in code `{s}`") # pyLambParse
    if len(m.body) > 1: raise JSTranspileError(f"py2js() function can only work with relatively simple operations. The code passed in is too complicated: `{s}`") # pyLambParse
    expr = m.body[0].value                                                       # pyLambParse
    if isinstance(expr, ast.Lambda): expr = expr.body                            # pyLambParse
    return expr                                                                  # pyLambParse
def py2js(m):                                                                    # py2js
    """Converts simple python expression into js code.
Example::

    kast.py2js("x+3")                 # returns ('(x+3)', {'x'})
    kast.py2js("lambda x: x+3")       # returns ('(x+3)', {'x'})
    kast.py2js("lambda y: y+3")       # returns ('(y+3)', {'y'})
    kast.py2js("lambda y: y+3*x")     # returns ('(y+(3*x))', {'x', 'y'})
    kast.py2js("x<3 and 45>10")       # returns ('((x<3)&&(45>10))', {'x'})
    kast.py2js("abs((x//5.1) * 2.3)") # returns ('Math.abs((Math.floor(x/5.1)*2.3))', {'x'})
    kast.py2js("not x > 3")           # returns ('!(x>3)', {'x'})

This will return a tuple of 2 objects. First is the JS code, and second is the
set of unknown variables in the JS code. If the code is a lambda function,
then it only considers what's inside the function. More complex examples::

    # returns ('[...Array(10).keys()].map((x, i) => [x, i]).filter(([x, y]) => ((x%3)===1)).map(([x, y]) => (Math.pow(x, y)+z))', {'z'})
    kast.py2js("[x**y + z for x, y in enumerate(range(10)) if x % 3 == 1]")

Both JS and Python code when run should return [1, 256, 823543] for z = 0.

A lot of builtin Python functions are transpiled automatically, including::

    abs, all, any, divmod, enumerate, float, getattr, hasattr, int, iter
    len, list, max, min, pow, print, range, reversed, round, set, setattr
    str, sum, tuple, type, zip

Other builtin operations don't quite make sense though. But the neat thing
is, any functions beyond these builtin functions are not transpiled, and
instead just passes along to the JS code. This allows you to kinda mix and
match code styles from both Python and JS. For example::

    # returns ("(document.querySelector(('#' + someIdx + '_tail')).style['color'] = (((Math.round(Math.pow(x, 2))===Math.abs(-12))) ? ('red') : ('green')))", {'someIdx', 'x'})
    kast.py2js("setattr(document.querySelector(f'#{someIdx}_tail').style, 'color', 'red' if round(x**2) == Math.abs(-12) else 'green')")

See how we're both using the Python transpiled ``**`` (to ``Math.pow(...)``),
yet uses ``Math.abs(...)`` directly, which is not defined in Python, but is
defined in JS, without any trouble. Also notice that ``document.querySelector``
and f-strings work too. So in short, overall syntax is Python's, but you can
call random JS functions!

This can do lots of transformations, but don't expect it to be airtight. I've
done all I can to make the thing airtight, but I'm sure you can come up with
some complex scenario where it would fail. This is meant to be applied on
short pieces of code only."""                                                    # py2js
    lcs = set(); ans = _py2js(m, lcs); return ans, lcs                           # py2js
def _listCompTarget(m, lcs): # extracts target of list comprehensions, like [... for x, [y, z] in ...] # _listCompTarget
    if isinstance(m, ast.Name): return _py2js(m, lcs)                            # _listCompTarget
    if isinstance(m, (ast.Tuple, ast.List)): s = ", ".join([_listCompTarget(e, lcs) for e in m.elts]); return f"[{s}]" # _listCompTarget
    raise JSTranspileError(f"While trying to expand target assignment in list/set/dict comprehensions, encountered unknown type: {type(m)}") # _listCompTarget
def _py2js(m, lcs:"list[str]"): # lcs for list of variable names, only used in ast.Name # _py2js
    if isinstance(m, str): m = pyLambParse(m)                                    # _py2js
    if isinstance(m, ast.BinOp):                                                 # _py2js
        left = _py2js(m.left, lcs); right = _py2js(m.right, lcs)                 # _py2js
        if isinstance(m.op, ast.FloorDiv): return f"Math.floor({left}/{right})" # TODO: might be weird dealing with negative floors # _py2js
        if isinstance(m.op, ast.Pow): return f"Math.pow({left}, {right})"        # _py2js
        return f"({left}{_py2js(m.op, lcs)}{right})"                             # _py2js
    if isinstance(m, ast.Compare): # only supports a<b or a<b<c, dont support more elems, like a<b<c<d # _py2js
        nops = len(m.ops); ncomps = len(m.comparators)                           # _py2js
        if nops != ncomps: raise JSTranspileError(f"Number of operations {nops} differed from number of comparators {ncomps}") # _py2js
        if nops == 0: raise JSTranspileError("Number of operations can't be zero") # _py2js
        if nops > 2: raise JSTranspileError("Don't support Compares with more than 2 comparisons. Change `a<b<c<d` into `(a<b<c) and (c<d)`") # _py2js
        left = _py2js(m.left, lcs); comps = [_py2js(e, lcs) for e in m.comparators]; ops = [_py2js(o, lcs) for o in m.ops] # _py2js
        if nops == 1: return f"({left}{ops[0]}{comps[0]})"                       # _py2js
        if nops == 2: return f"({left}{ops[0]}{comps[0]}) && ({comps[0]}{ops[1]}{comps[1]})" # _py2js
        return m                                                                 # _py2js
    if isinstance(m, ast.BoolOp): return f"({_py2js(m.op, lcs).join([_py2js(v, lcs) for v in m.values])})" # _py2js
    if isinstance(m, ast.UnaryOp):                                               # _py2js
        if isinstance(m.op, ast.Not): return f"!{_py2js(m.operand, lcs)}"        # _py2js
        if isinstance(m.op, ast.Invert): raise JSTranspileError(f"No notion of inversion '~' exists in JS") # _py2js
        if isinstance(m.op, ast.USub): return f"-{_py2js(m.operand, lcs)}"       # _py2js
        if isinstance(m.op, ast.UAdd): return f"{_py2js(m.operand, lcs)}"        # _py2js
        raise JSTranspileError(f"kast.py2js() doesn't know how to parse UnaryOp `{m.op}`") # _py2js
    if isinstance(m, ast.Name): lcs.add(m.id); return f"{m.id}"                  # _py2js
    if isinstance(m, ast.Constant):                                              # _py2js
        if m.value is None: return f"null"                                       # _py2js
        if isinstance(m.value, str): s = m.value.replace("'", "\\'"); return f"'{s}'" # _py2js
        return f"{m.value}"                                                      # _py2js
    if isinstance(m, ast.Add): return f"+"                                       # _py2js
    if isinstance(m, ast.Sub): return f"-"                                       # _py2js
    if isinstance(m, ast.Mult): return f"*"                                      # _py2js
    if isinstance(m, ast.MatMult): raise JSTranspileError("No notion of matrix multiplication '@' exists in JS") # _py2js
    if isinstance(m, ast.Div): return f"/"                                       # _py2js
    if isinstance(m, ast.Mod): return f"%"                                       # _py2js
    if isinstance(m, ast.Pow): return f"^"                                       # _py2js
    if isinstance(m, ast.Eq): return f"==="                                      # _py2js
    if isinstance(m, ast.NotEq): return f"!=="                                   # _py2js
    if isinstance(m, ast.Gt): return f">"                                        # _py2js
    if isinstance(m, ast.GtE): return f">="                                      # _py2js
    if isinstance(m, ast.Lt): return f"<"                                        # _py2js
    if isinstance(m, ast.LtE): return f"<="                                      # _py2js
    if isinstance(m, ast.LShift): return "<<"                                    # _py2js
    if isinstance(m, ast.RShift): return ">>"                                    # _py2js
    if isinstance(m, ast.And): return "&&"                                       # _py2js
    if isinstance(m, ast.Or): return "||"                                        # _py2js
    if isinstance(m, ast.Attribute): return f"{_py2js(m.value, lcs)}.{m.attr}"   # _py2js
    if isinstance(m, ast.Call):                                                  # _py2js
        if len(m.keywords) > 0: raise JSTranspileError(f"keyword arguments don't exist in JS, please use normal arguments in your function calls only, at `{m}`") # _py2js
        funcName = flattenAttribute(m.func)                                      # _py2js
        if funcName in jstCallDict: return jstCallDict[funcName](m, lcs)         # _py2js
        return f"{funcName}({', '.join([_py2js(a, lcs) for a in m.args])})"      # _py2js
    if isinstance(m, (ast.List, ast.Tuple)): return f"[{', '.join([_py2js(e, lcs) for e in m.elts])}]" # _py2js
    if isinstance(m, ast.Set): return f"new Set([{', '.join([_py2js(e, lcs) for e in m.elts])}])" # _py2js
    if isinstance(m, ast.Dict):                                                  # _py2js
        keys = [f"{_py2js(k, lcs)}" if isinstance(k, ast.Constant) else (f"[{_py2js(k, lcs)}]" if k is not None else None) for k in m.keys] # _py2js
        values = [f"{_py2js(v, lcs)}" for v in m.values]                         # _py2js
        a = [f"{k}: {v}" for k, v in zip(keys, values) if k is not None]         # _py2js
        b = [f"...{v}" for k, v in zip(keys, values) if k is None]               # _py2js
        return "{" + ", ".join([*a, *b]) + "}"                                   # _py2js
    if isinstance(m, ast.Starred): return f"...{_py2js(m.value, lcs)}"           # _py2js
    if isinstance(m, ast.IfExp): return f"(({_py2js(m.test, lcs)}) ? ({_py2js(m.body, lcs)}) : ({_py2js(m.orelse, lcs)}))" # _py2js
    if isinstance(m, ast.Subscript):                                             # _py2js
        tail = ""; slices = m.slice.elts if isinstance(m.slice, ast.Tuple) else [m.slice] # _py2js
        for s in reversed(slices):                                               # _py2js
            if isinstance(s, ast.Slice):                                         # _py2js
                if s.step is not None: pr = '\n'.join(m | pretty()); raise JSTranspileError(f"JS does not support stepped slices! This can be augmented on my side later on though. kast.pretty() on the element:\n\n{pr}") # _py2js
                l = 'undefined' if s.lower is None else _py2js(s.lower, lcs)     # _py2js
                u = 'undefined' if s.upper is None else _py2js(s.upper, lcs)     # _py2js
                if tail: dataIdx = k1lib.cli.init._jsDAuto(); tail = f".slice({l}, {u}).map(({dataIdx}) => {dataIdx}{tail})" # _py2js
                else: tail = f".slice({l}, {u})" # short path                    # _py2js
            else: tail = f".at({_py2js(s, lcs)}){tail}"                          # _py2js
        return f"{_py2js(m.value, lcs)}{tail}"                                   # _py2js
    if isinstance(m, (ast.ListComp, ast.GeneratorExp, ast.SetComp)):             # _py2js
        if len(m.generators) != 1: raise JSTranspileError("Currently only supports 1 generator in list/set/generator comprehensions") # _py2js
        g = m.generators[0]; targetLcs = set(); oldLcs = set(lcs)                # _py2js
        if g.is_async: raise JSTranspileError("Does not support async functions yet") # _py2js
        ans = _py2js(g.iter, lcs); target = _listCompTarget(g.target, targetLcs) # _py2js
        for if_ in g.ifs: ans = f"{ans}.filter(({target}) => {_py2js(if_, lcs)})" # _py2js
        ans = f"{ans}.map(({target}) => {_py2js(m.elt, lcs)})"                   # _py2js
        for x in list(lcs): # try removing unknown variables in the .elt expression that appears in the target, cause those are not unknown variables! # _py2js
            if x not in oldLcs and x in targetLcs: lcs.remove(x)                 # _py2js
        return f"new Set({ans})" if isinstance(m, ast.SetComp) else ans          # _py2js
    if isinstance(m, ast.DictComp):                                              # _py2js
        if len(m.generators) != 1: raise JSTranspileError("Currently only supports 1 generator in dict comprehensions") # _py2js
        g = m.generators[0]; targetLcs = set(); oldLcs = set(lcs)                # _py2js
        if g.is_async: raise JSTranspileError("Does not support async functions yet") # _py2js
        ans = _py2js(g.iter, lcs); target = _listCompTarget(g.target, targetLcs) # _py2js
        for if_ in g.ifs: ans = f"{ans}.filter(({target}) => {_py2js(if_, lcs)})" # _py2js
        ans = f"{ans}.map(({target}) => [{_py2js(m.key, lcs)}, {_py2js(m.value, lcs)}]).toDict()" # _py2js
        for x in list(lcs): # try removing unknown variables in the .elt expression that appears in the target, cause those are not unknown variables! # _py2js
            if x not in oldLcs and x in targetLcs: lcs.remove(x)                 # _py2js
        return ans                                                               # _py2js
    if isinstance(m, ast.FormattedValue): return _py2js(m.value, lcs)            # _py2js
    if isinstance(m, ast.JoinedStr): return f"({' + '.join([_py2js(e, lcs) for e in m.values])})" # _py2js
    raise JSTranspileError(f"kast.py2js() doesn't know how to parse {type(m)}")  # _py2js
def prepareFunc(fn:"str|cli.op", extraVars:"list[str]"=None, checks=True):       # prepareFunc
    """for toJsFunc() transpilation subsystem.
Returns (transpiled function, variable declarations, argument variables).
If can't understand the object passed in, return None

:param fn: assumes this is a cli.op(), a lambda function, or an expression inside a lambda function
:param extraVars: available variables that can be resolved from js
    function. Can come from arguments, or from js function declaration,
    or other places. Specify this to shut up the variable resolver checks"""     # prepareFunc
    g = cli.init._k1_global_frame(); avaiVals = dict(); reqVals = set(); extraVars = extraVars or [] # prepareFunc
    if isinstance(fn, cli.op):                                                   # prepareFunc
        fn, vs = fn.ab_fastFS()                                                  # prepareFunc
        for v, vv in vs.items(): reqVals.add(v); avaiVals[v] = vv                # prepareFunc
    elif not isinstance(fn, str): return None; raise Exception(f"Function inside apply() has to either be a cli.op(), or a string containg a Python expression. Instead, it's {type(fn)}") # prepareFunc
    # adds args from lambda func to available vars                               # prepareFunc
    argVars = kast_lambda(fn); extraVars.extend(argVars)                         # prepareFunc
    # transpiles to js                                                           # prepareFunc
    fn, vs = py2js(fn)                                                           # prepareFunc
    # try to resolve variables                                                   # prepareFunc
    for v in vs:                                                                 # prepareFunc
        reqVals.add(v)                                                           # prepareFunc
        if v == "x": continue # just ignore "x" variable, cause that's the default # prepareFunc
        try: avaiVals[v] = g[v]                                                  # prepareFunc
        except: pass                                                             # prepareFunc
    # removing available variables from required variables                       # prepareFunc
    for v in avaiVals: reqVals.remove(v)                                         # prepareFunc
    extraVars = set(extraVars)                                                   # prepareFunc
    if checks:                                                                   # prepareFunc
        for v in reqVals:                                                        # prepareFunc
            if v not in extraVars:                                               # prepareFunc
                raise Exception(f"Expression depends on the variables {v}, but they couldn't be resolved") # prepareFunc
    # forming declaration statements                                             # prepareFunc
    return fn, "\n".join([f"const {k} = {json.dumps(v)};" for k, v in avaiVals.items()]), argVars # prepareFunc
def kast_lambda(fn:"str|cli.op") -> "list[str]":                                 # kast_lambda
    """Grab args of the input lambda func. If it's just a simple
expression then return ['x']"""                                                  # kast_lambda
    l = ast.parse(fn).body[0].value                                              # kast_lambda
    if isinstance(l, ast.Lambda): return [a.arg for a in l.args.args]            # kast_lambda
    return ["x"]                                                                 # kast_lambda
def prepareFunc2(fn, checks=True) -> "(header, fIdx) | None":                    # prepareFunc2
    res = prepareFunc(fn, [], checks)                                            # prepareFunc2
    if res is None: return None                                                  # prepareFunc2
    expr, avaiVars, args = res                                                   # prepareFunc2
    fIdx = k1lib.cli.init._jsFAuto()                                             # prepareFunc2
    dataIdx = k1lib.cli.init._jsDAuto()                                          # prepareFunc2
    return f"""{avaiVars}
const {fIdx} = ({', '.join(args)}) => {{
    return {expr};
}}""", fIdx                                                                      # prepareFunc2
def prepareFunc3(fn, meta, kwargs=None, args=None) -> "(header, fIdx)":          # prepareFunc3
    """Kinda similar to prepareFunc2(), but this time uses 3 different
strategies to transpile:

- Using prepareFunc2(), works for string with lambda functions and cli.op()
- Using the object's builtin ._jsF() function
- Using k1lib.settings.cli.kjs.jsF dictionary of different custom _jsF() functions

Throws error if can't transpile the function

:param kwargs: keyword arguments and arguments provided by aS(), apply(), filt(), and so on.
"""                                                                              # prepareFunc3
    res = prepareFunc2(fn, False) # tries to compile str/cli.op                  # prepareFunc3
    if res:                                                                      # prepareFunc3
        if not(kwargs is None or len(kwargs) == 0): raise Exception(f"Keyword arguments doesn't exist in JS! Please convert the function `{fn}` into positional arguments") # prepareFunc3
        if args is None or len(args) == 0: return res                            # prepareFunc3
        # prepare extra code here if extra args are available                    # prepareFunc3
        fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto(); argIdx = cli.init._jsDAuto() # prepareFunc3
        _header, _fIdx = res; return f"""{_header}\nconst {argIdx} = {json.dumps(args)};\nconst {fIdx} = ({dataIdx}) => {_fIdx}({dataIdx}, ...{argIdx})""", fIdx # prepareFunc3
    elif hasattr(fn, "_jsF"): return fn._jsF(meta, *(args or ()), **(kwargs or {})) # if not str/cli.op, then lookup the funcs, then pass through the kwargs and whatnot # prepareFunc3
    elif fn in k1lib.settings.cli.kjs.jsF:                                       # prepareFunc3
        try: return k1lib.settings.cli.kjs.jsF[fn](meta, *(args or ()), **(kwargs or {})) # prepareFunc3
        except Exception as e: print(f"Function `{fn}` exist in system, but throws the below error when executed"); raise e # prepareFunc3
    else: raise Exception(f"*._jsF() right now doesn't know how to transpile to JS this func: {fn}\n\nIt could be you're using regular lambda functions. Convert it to use cli.op(), or just enter the lambda function as a string, so `lambda x: x**2` turns into the string `\"lambda x: x**2\"`\n\nOr, if it's a function you use often, add to `settings.cli.kjs.jsF`") # prepareFunc3
def asyncGuard(data:"(header, fIdx)") -> "(header, fIdx, async?)":               # asyncGuard
    """Converts ._jsF() function signatures from (header, fIdx) into (header, fIdx, is fIdx async?).
Example::

    # returns tuple (header, fIdx, async?)
    kast.asyncGuard(apply("x**2")._jsF(None))
"""                                                                              # asyncGuard
    if data is NotImplemented: return NotImplemented                             # asyncGuard
    header, fIdx = data                                                          # asyncGuard
    g1 = cli.grep(f"(?P<a>(const)|(let)) {fIdx}[ ]*=[ ]*(?P<g>(async)|((?!\(*)))[ ]*\(", extract="g") # asyncGuard
    g2 = cli.grep(f"(?P<g>async)[ ]*function[ ]*{fIdx}\(", extract="g")          # asyncGuard
    x = header.split("\n"); return [header, fIdx, len(list(g1(x))) + len(list(g2(x))) > 0] # asyncGuard



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This includes helper clis that make it quick to graph graphviz plots."""
__all__ = ["sketch", "edges"]
import re, k1lib, math, os, numpy as np, io, json, base64, unicodedata, inspect
from k1lib.cli.init import BaseCli; import k1lib.cli as cli, k1lib.cli.init as init
graphviz = k1lib.dep("graphviz")
from collections import deque, defaultdict
settings = k1lib.settings.cli
_svgAutoInc = k1lib.AutoIncrement(prefix="_k1_svg_")
_preAutoInc = k1lib.AutoIncrement(prefix="_k1_svg_pre_")
_clusterAuto = k1lib.AutoIncrement(prefix="cluster_")
_idxAuto = k1lib.AutoIncrement(prefix="_idx_")
class sketch(BaseCli):                                                           # sketch
    _g = None; _name2Idx = None; ctxIdx = None                                   # sketch
    def __init__(self, **kwargs):                                                # sketch
        """Similar to :class:`~utils.sketch`, which makes it easier to plot multiple graphs quickly,
this makes it easier to plot node graphs a lot quicker than I have before. This cli
configures the graph in general, but doesn't dive too much into the specifics. Example::

    ["ab", "bc", "ca"] | (kgv.sketch(engine="sfdp") | kgv.edges())

Most of the complexities are in :class:`edges`, so check that class out for more comprehensive examples

:param kwargs: keyword arguments passed into :class:`graphviz.Digraph`"""        # sketch
        super().__init__(capture=True); self.kwargs = kwargs                     # sketch
    @staticmethod                                                                # sketch
    def _guard():                                                                # sketch
        if sketch._g is None: raise Exception("Context has not been setup yet, can't proceed to plot the edges/nodes. This could be because you're doing `data | kgv.edges()` directly. Instead, do `data | (kgv.sketch() | kgv.edges())`. The operation `kgv.sketch()` will initialize the graph") # sketch
    def __ror__(self, it):                                                       # sketch
        sketch._g = g = graphviz.Digraph(**self.kwargs); sketch._idx2Popup = idx2Popup = {} # sketch
        sketch._name2Idx = name2Idx = defaultdict(lambda: defaultdict(lambda: _idxAuto())) # sketch
        it | self.capturedSerial | cli.deref()                                   # sketch
        sketch._g = None; sketch._idx2Popup = None; sketch._name2Idx = None; return Graph(g, name2Idx, idx2Popup) # sketch
    def _jsF(self, meta):                                                        # sketch
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); ctxIdx = init._jsDAuto(); sketch.ctxIdx = ctxIdx # sketch
        header, _fIdx, _async = k1lib.kast.asyncGuard(self.capturedSerial._jsF(meta)) # sketch
        res = f"""\
let {ctxIdx} = null;\n{header}
const {fIdx} = async ({dataIdx}) => {{
    {ctxIdx} = [];
    const out = {'await ' if _async else ''}{_fIdx}({dataIdx});

    const res = (await (await fetch("https://local.mlexps.com/routeServer/kapi_10-graphviz", {{
      method: "POST",
      body: JSON.stringify({{ "obj": {ctxIdx} }}),
      headers: {{ "Content-Type": "application/json" }}
    }})).json());
    if (!res.success) throw new Error(res.reason);
    return atob(res.data);
}}""", fIdx                                                                      # sketch
        sketch.ctxIdx = None; return res                                         # sketch
class nodes(BaseCli):                                                            # nodes
    def __init__(self):                                                          # nodes
        pass                                                                     # nodes
    def __ror__(self, _it) -> "5-column input":                                  # nodes
        if sketch._g is None: return _it | (sketch() | self)                     # nodes
        sketch._guard(); it = []                                                 # nodes
        for row in _it:                                                          # nodes
            n = len(row)                                                         # nodes
            if n == 1: it.append(["", row[0], {}])                               # nodes
            elif n == 2: it.append([*row, {}])                                   # nodes
            elif n == 3: it.append(row)                                          # nodes
            else: raise Exception(f"kgv.nodes() can only accept tables from 1 to 3 columns. Detected {n} columns instead") # nodes
        g = sketch._g; name2Idx = sketch._name2Idx                               # nodes
        for s1,n1,kw in it:                                                      # nodes
            idx = name2Idx[s1][n1]                                               # nodes
            if "popup" in kw: sketch._idx2Popup[idx] = kw["popup"]; kw = dict(kw); del kw["popup"] # nodes
            g.node(idx, **kw)                                                    # nodes
        return it                                                                # nodes
    def _jsF(self, meta):                                                        # nodes
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); ctxIdx = sketch.ctxIdx; # nodes
        if ctxIdx is None: return (sketch() | self)._jsF(meta)                   # nodes
        return f"""\
const {fIdx} = ({dataIdx}) => {{
    const it = [];
    for (const row of {dataIdx}) {{
        const n = row.length;
        if (n === 1) it.push(["", row[0], {{}}])
        else if (n === 2) it.push([...row, {{}}])
        else if (n === 3) it.push(row)
        else throw new Error(`kgv.nodes() can only accept tables from 1 to 3 columns. Detected ${{n}} columns instead`)
    }}
    {ctxIdx}.push(["nodes", {{it, args: []}}]); return it;
}}
""", fIdx                                                                        # nodes
def drawSimple(g, names, data, name2Idx): # data here is List[[name1, name2, kw]], names is List[name] # drawSimple
    for name in set(names): g.node(name2Idx[name], name);                        # drawSimple
    for name1, name2, kw in data: g.edge(name2Idx[name1], name2Idx[name2], **kw) # drawSimple
class edges(BaseCli):                                                            # edges
    def __init__(self):                                                          # edges
        """Plots out edges of the graph.
Example 1::

    ["ab", "bc", "ca"] | kgv.edges()

Result:

.. image:: ../images/kgv_edges_1.png

If you need to customize the graph on initialization, then you can use :class:`sketch` to
capture related operations (like :class:`edges`), and inject your params in :class:`sketch`::

    ["ab", "bc", "ca"] | (kgv.sketch(engine="sfdp") | kgv.edges())

Example 2::

    [["a", "b", {"label": "3%"}], ["b", "c", {"color": "red"}], "ac", "cb"] | kgv.edges()

Result:

.. image:: ../images/kgv_edges_2.png

Example 3::

    [["group1", "a", "group1", "b", {"label": "3%"}],
     "ec",
     ["group1", "b", "", "c", {"color": "red"}],
     ["group1", "a", "", "c"],
     ["", "c", "group1", "b"],
     ["", "c", "group2", "d", {"color": "green"}]
    ] | kgv.edges()

Result:

.. image:: ../images/kgv_edges_3.png

So the idea is, each row describes a single edge on the graph. Each row
can be multiple different lengths, but only these configurations are allowed:

- [name1, name2]
- [name1, name2, kwargs]
- [group1, name1, group2, name2]
- [group1, name1, group2, name2, kwargs]

So if you don't need the complexity and just want to plot something out, you
can just use the one at the top, but if you do want fancy features, then you
can add those in the kwargs.

Check out a gallery of more examples at `kapi/10-graphviz <https://mlexps.com/kapi/10-graphviz/>`_.
"""                                                                              # edges
        pass                                                                     # edges
    def __ror__(self, _it) -> "5-column input":                                  # edges
        if sketch._g is None: return _it | (sketch() | self)                     # edges
        sketch._guard(); it = []                                                 # edges
        for row in _it:                                                          # edges
            n = len(row)                                                         # edges
            if n == 2: it.append(["", row[0], "", row[1], {}])                   # edges
            elif n == 3: it.append(["", row[0], "", row[1], row[2]])             # edges
            elif n == 4: it.append([*row, {}])                                   # edges
            elif n == 5: it.append(row)                                          # edges
            else: raise Exception(f"kgv.edges() can only accept tables from 2 to 5 columns. Detected {n} columns instead") # edges
        g = sketch._g; name2Idx = sketch._name2Idx                               # edges
        # grouping by segments and drawing their internals first                 # edges
        for segN, names in it | cli.batched(2).all() | cli.joinSt() | cli.groupBy(0, True) | cli.apply(cli.joinSt(), 1) | cli.deref(): # edges
            if segN:                                                             # edges
                with g.subgraph(name=_clusterAuto()) as subG:                    # edges
                    subG.attr(label=f"{segN}"); drawSimple(subG, names, it | cli.filt(cli.op() == segN, [0, 2]) | cli.cut(1, 3, 4) | cli.deref(), name2Idx[segN]) # edges
            else: drawSimple(g, names, it | cli.filt(cli.op() == segN, [0, 2]) | cli.cut(1, 3, 4) | cli.deref(), name2Idx[""]) # edges
        # then draw external edges                                               # edges
        for s1, n1, s2, n2, kw in it | cli.filt(lambda x: x[0] != x[2]): g.edge(name2Idx[s1][n1], name2Idx[s2][n2], **kw) # edges
        return it                                                                # edges
    def _jsF(self, meta):                                                        # edges
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); ctxIdx = sketch.ctxIdx; # edges
        if ctxIdx is None: return (sketch() | self)._jsF(meta)                   # edges
        return f"""\
const {fIdx} = ({dataIdx}) => {{
    // why not just pass dataIdx in directly? Well, in Python, the interface is that __ror__ should return a 5-column input, so here, gotta honor that, in case the user has some operation downstream of this
    const it = [];
    for (const row of {dataIdx}) {{
        const n = row.length;
        if (n === 2) it.push(["", row[0], "", row[1], {{}}])
        else if (n === 3) it.push(["", row[0], "", row[1], row[2]])
        else if (n === 4) it.push([...row, {{}}])
        else if (n === 5) it.push(row)
        else throw new Error(`kgv.edges() can only accept tables from 2 to 5 columns. Detected ${{n}} columns instead`)
    }}
    {ctxIdx}.push(["edges", {{it, args: []}}]); return it;
}}
""", fIdx                                                                        # edges
class Graph:                                                                     # Graph
    def __init__(self, g, name2Idx, idx2Popup):                                  # Graph
        """Wrapper around a :class:`graphviz.Graph` or :class:`graphviz.Digraph`. Internal graph
object is available at ``self.g``. Not instantiated by end user, instead, returned by :class:`sketch`.""" # Graph
        self.g = g; self.name2Idx = name2Idx; self.idx2Popup = idx2Popup         # Graph
    def _repr_mimebundle_(self, *args, **kwargs): return self.g._repr_mimebundle_(*args, **kwargs) # Graph
    def _repr_html_(self): return self._toHtml()                                 # Graph
    def _toHtml(self):                                                           # Graph
        idx2Popup = self.idx2Popup; name2Idx = self.name2Idx; s = self.g | cli.toHtml() | cli.op().replace("><", ">\n<") # Graph
        # these ids are: List[(auto generated id, unique id we're replacing it with)] # Graph
        # nodeIds = s.split("\n") | grep('<g id="(?P<g>node[0-9]+)"', extract="g") | apply(lambda x: [x, _svgAutoInc()]) | deref() # Graph
        nodeIds = s.split("\n") | cli.grep('class="node"', after=1) | cli.batched(2) | cli.apply(cli.op().split("title>")[1].strip("</"), 1) | cli.grep('<g id="(?P<g>node[0-9]+)', col=0, extract="g") | cli.deref() # Graph
        edgeIds = s.split("\n") | cli.grep('<g id="(?P<g>edge[0-9]+)"', extract="g") | cli.apply(lambda x: [x, _svgAutoInc()]) | cli.deref() # Graph
        graphIds = s.split("\n") | cli.grep('<g id="(?P<g>graph[0-9]+)"', extract="g") | cli.apply(lambda x: [x, _svgAutoInc()]) | cli.deref() # Graph
        for x, y in [nodeIds, edgeIds, graphIds] | cli.joinSt(): s = s.replace(f'id="{x}"', f'id="{y}"') # Graph
        a = nodeIds | cli.cut(1) | cli.apply(lambda idx: [idx, idx2Popup.get(idx, None)]) | cli.deref(); pre = _preAutoInc() # Graph
        inside = f"rect.x <= {pre}_mouseX && {pre}_mouseX < rect.x+rect.width && rect.y <= {pre}_mouseY && {pre}_mouseY < rect.y+rect.height" # Graph
        return f"""
<div id="{pre}_wrapper" style="position: relative">{s}<div id="{pre}_popup" style="position: absolute; display: none; background: white; padding: 8px 12px; border-radius: 6px; box-shadow: 0 3px 5px rgb(0 0 0 / 0.3); z-index: 1000000"></div></div>
<script>
    const {pre}_nodeId_node_popup = ({json.dumps(a)}).map(([x,y]) => [x, document.querySelector(`#${{x}}`), y]);
    const {pre}_nodes = {pre}_nodeId_node_popup.map(([x,n,y]) => n); let {pre}_activeNode = null;
    const {pre}_popup = document.querySelector("#{pre}_popup"); const {pre}_wrapper = document.querySelector("#{pre}_wrapper");
    const {pre}_nodeId2popup = {{}}; for (const [x,n,y] of {pre}_nodeId_node_popup) {{ {pre}_nodeId2popup[x] = y; }};
    let {pre}_mouseX = 0; let {pre}_mouseY = 0; {pre}_wrapper.onmousemove = (e) => {{ {pre}_mouseX = e.clientX; {pre}_mouseY = e.clientY; }};

    setInterval(() => {{
        if ({pre}_activeNode) {{
            const rect = {pre}_activeNode.getBoundingClientRect();
            if (!({inside})) {{ {pre}_activeNode = null; {pre}_popup.innerHTML = ""; {pre}_popup.style.display = "none"; }}
        }}

        if (!{pre}_activeNode) {{ // can't just do `if (activeNode) ... else ...` btw. Separated out for a reason
            const wRect = {pre}_wrapper.getBoundingClientRect();
            for (const node of {pre}_nodes) {{
                const rect = node.getBoundingClientRect();
                if ({inside}) {{
                    const popup = {pre}_nodeId2popup[node.id];
                    {pre}_activeNode = node;
                    if (popup) {{
                        {pre}_popup.style.left = rect.x + rect.width/2 + 10 - wRect.x + "px";
                        {pre}_popup.style.top = 0; {pre}_popup.innerHTML = popup; {pre}_popup.style.display = "block";
                        const pRect = {pre}_popup.getBoundingClientRect();
                        const t1 = rect.y + rect.height/2 + 10 - wRect.y; // "t" for "top"
                        const t2 = wRect.height - pRect.height;
                        {pre}_popup.style.top = ((t2 < 0) ? 0 : Math.min(t1, t2)) + "px";
                    }}
                    break;
                }}
            }}
        }}
    }}, 30);
</script>"""                                                                     # Graph
        return s                                                                 # Graph



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
.. module:: k1lib
"""
from typing import Callable, Iterator, Tuple, Union, Dict, Any, List
from k1lib import isNumeric; import k1lib, contextlib, warnings
import random, math, sys, io, os, numpy as np, functools
plt = k1lib.dep("matplotlib.pyplot")
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["Object", "Range", "Domain", "AutoIncrement", "Wrapper", "Every",
           "RunOnce", "MaxDepth", "MovingAvg", "Absorber",
           "Settings", "settings", "_settings", "UValue", "ConstantPad"]
class Object:                                                                    # Object
    """Convenience class that acts like :class:`~collections.defaultdict`. You
can use it like a normal object::

    a = k1lib.Object()
    a.b = 3
    print(a.b) # outputs "3"

``__repr__()`` output is pretty nice too:

.. code-block:: text

    <class '__main__.Object'>, with attrs:
    - b

You can instantiate it from a dict::

    a = k1lib.Object.fromDict({"b": 3, "c": 4})
    print(a.c) # outputs "4"

And you can specify a default value, just like defaultdict::

    a = k1lib.Object().withAutoDeclare(lambda: [])
    a.texts.extend(["factorio", "world of warcraft"])
    print(a.texts[0]) # outputs "factorio"

.. warning::

    Default values only work with variables that don't start with an
    underscore "_".

Treating it like defaultdict is okay too::

    a = k1lib.Object().withAutoDeclare(lambda: [])
    a["movies"].append("dune")
    print(a.movies[0]) # outputs "dune" """                                      # Object
    def __init__(self): self._defaultValueGenerator = None; self.repr = None     # Object
    @staticmethod                                                                # Object
    def fromDict(_dict:Dict[str, Any]):                                          # Object
        """Creates an object with attributes from a dictionary"""                # Object
        answer = Object(); answer.__dict__.update(_dict); return answer          # Object
    @property                                                                    # Object
    def state(self) -> dict:                                                     # Object
        """Essentially ``__dict__``, but only outputs the fields you
defined. If your framework intentionally set some attributes, those
will be reported too, so beware"""                                               # Object
        answer = dict(self.__dict__); del answer["_defaultValueGenerator"]       # Object
        del answer["repr"]; return answer                                        # Object
    def withAutoDeclare(self, defaultValueGenerator):                            # Object
        """Sets this Object up so that if a field doesn't
        exist, it will automatically create it with a
        default value."""                                                        # Object
        self._defaultValueGenerator = defaultValueGenerator; return self         # Object
    def __getitem__(self, idx): return getattr(self, idx)                        # Object
    def __setitem__(self, idx, value): setattr(self, idx, value)                 # Object
    def __iter__(self): yield from self.state.values()                           # Object
    def __contains__(self, item:str): return item in self.__dict__               # Object
    def __getattr__(self, attr):                                                 # Object
        if attr.startswith("_"): raise AttributeError()                          # Object
        if attr == "getdoc": raise AttributeError("This param is used internally in module `IPython.core.oinspect`, so you kinda have to set it specifically yourself instead of relying on auto declare") # Object
        if self._defaultValueGenerator != None:                                  # Object
            self.__dict__[attr] = self._defaultValueGenerator()                  # Object
            return self.__dict__[attr]                                           # Object
        raise AttributeError                                                     # Object
    def __delitem__(self, key): del self.__dict__[key]                           # Object
    def withRepr(self, _repr:str):                                               # Object
        """Specify output of ``__repr__()``. Legacy code. You can just
monkey patch it instead."""                                                      # Object
        self.repr = _repr; return self                                           # Object
    def __repr__(self):                                                          # Object
        _dict = "\n".join([f"- {k}" for k in self.state.keys()])                 # Object
        return self.repr or f"{type(self)}, with attrs:\n{_dict}"                # Object
ninf = float("-inf"); inf = float("inf")                                         # Object
class Range:                                                                     # Range
    """A range of numbers. It's just 2 numbers really: start and stop

This is essentially a convenience class to provide a nice, clean
abstraction and to eliminate errors. You can transform values::

    Range(10, 20).toUnit(13) # returns 0.3
    Range(10, 20).fromUnit(0.3) # returns 13
    Range(10, 20).toRange(Range(20, 10), 13) # returns 17

You can also do random math operations on it::

    (Range(10, 20) * 2 + 3) == Range(23, 43) # returns True
    Range(10, 20) == ~Range(20, 10) # returns True"""                            # Range
    def __init__(self, start=0, stop=None):                                      # Range
        """Creates a new Range.

There are different ``__init__`` functions for many situations:

- Range(2, 11.1): create range [2, 11.1]
- Range(15.2): creates range [0, 15.2]
- Range(Range(2, 3)): create range [2, 3]. This serves as sort of a catch-all
- Range(slice(2, 5, 2)): creates range [2, 5]. Can also be a :class:`range`
- Range(slice(2, -1), 10): creates range [2, 9]
- Range([1, 2, 7, 5]): creates range [1, 5]. Can also be a tuple
"""                                                                              # Range
        if (isNumeric(start) and isNumeric(stop)):                               # Range
            self.start, self.stop = start, stop                                  # Range
        elif isNumeric(start) and stop == None:                                  # Range
            self.start, self.stop = 0, start                                     # Range
        elif stop == None and isinstance(start, (range, slice, Range)):          # Range
            self.start, self.stop = start.start, start.stop                      # Range
        elif isNumeric(stop) and isinstance(start, slice):                       # Range
            r = range(stop)[start]; self.start, self.stop = r.start, r.stop      # Range
        elif isinstance(start, (list, tuple)):                                   # Range
            self.start, self.stop = start[0], start[-1]                          # Range
        else: raise AttributeError(f"Don't understand {start} and {stop}")       # Range
        self.delta = self.stop - self.start                                      # Range
    def __getitem__(self, index):                                                # Range
        """0 for start, 1 for stop

You can also pass in a :class:`slice` object, in which case, a range subset
will be returned. Code kinda looks like this::

    range(start, stop)[index]"""                                                 # Range
        if index == 0: return self.start                                         # Range
        if index == 1: return self.stop                                          # Range
        if type(index) == slice:                                                 # Range
            return Range(range(self.start, self.stop)[index])                    # Range
        raise Exception(f"Can't get index {index} of range [{self.start}, {self.stop}]") # Range
    def fixOrder(self) -> "Range":                                               # Range
        """If start greater than stop, switch the 2, else do nothing"""          # Range
        if self.start > self.stop:                                               # Range
            self.start, self.stop = self.stop, self.start                        # Range
        return self                                                              # Range
    def _common(self, x, f:Callable[[float], float]):                            # Range
        if isNumeric(x): return f(x)                                             # Range
        if isinstance(x, (list, tuple)):                                         # Range
            return [self._common(elem, f) for elem in x]                         # Range
        if isinstance(x, (range, slice, Range)):                                 # Range
            return Range(self._common(x.start if x.start != None else 0, f), self._common(x.stop if x.stop != None else 1, f)) # Range
        raise AttributeError(f"Doesn't understand {x}")                          # Range
    def __iter__(self): yield self.start; yield self.stop                        # Range
    def intIter(self, step:int=1) -> Iterator[int]:                              # Range
        """Returns integers within this Range"""                                 # Range
        return range(int(self.start), int(self.stop), step)                      # Range
    def toUnit(self, x):                                                         # Range
        """Converts x from current range to [0, 1] range. Example::

    r = Range(2, 10)
    r.toUnit(5) # will return 0.375, as that is (5-2)/(10-2)

You can actually pass in a lot in place of x::

    r = Range(0, 10)
    r.toUnit([5, 3, 6]) # will be [0.5, 0.3, 0.6]. Can also be a tuple
    r.toUnit(slice(5, 6)) # will be slice(0.5, 0.6). Can also be a range, or Range

.. note::

    In the last case, if ``start`` is None, it gets defaulted to 0, and
    if ``end`` is None, it gets defaulted to 1
"""                                                                              # Range
        def f(x):                                                                # Range
            if self.delta == 0: return float("nan")                              # Range
            return (x - self.start) / self.delta                                 # Range
        return self._common(x, lambda x: float("nan") if self.delta == 0 else (x - self.start) / self.delta) # Range
    def fromUnit(self, x):                                                       # Range
        """Converts x from [0, 1] range to this range. Example::

    r = Range(0, 10)
    r.fromUnit(0.3) # will return 3

x can be a lot of things, see :meth:`toUnit` for more"""                         # Range
        return self._common(x, lambda x: x * self.delta + self.start)            # Range
    def toRange(self, _range:"Range", x):                                        # Range
        """Converts x from current range to another range. Example::

    r = Range(0, 10)
    r.toRange(Range(0, 100), 6) # will return 60

x can be a lot of things, see :meth:`toUnit` for more."""                        # Range
        return self._common(x, lambda x: Range(_range).fromUnit(self.toUnit(x))) # Range
    def fromRange(self, _range:"Range", x):                                      # Range
        """Reverse of :meth:`toRange`, effectively."""                           # Range
        return _range.toRange(self, x)                                           # Range
    @property                                                                    # Range
    def range_(self):                                                            # Range
        """Returns a :class:`range` object with start and stop values
rounded off"""                                                                   # Range
        return range(math.floor(self.start+0.001), math.floor(self.stop+0.001))  # Range
    @property                                                                    # Range
    def slice_(self):                                                            # Range
        """Returns a :class:`slice` object with start and stop values
rounded off"""                                                                   # Range
        return slice(math.floor(self.start+0.001), math.floor(self.stop+0.001))  # Range
    @staticmethod                                                                # Range
    def proportionalSlice(r1, r2, r1Slice:slice) -> Tuple["Range", "Range"]:     # Range
        """Slices r1 and r2 proportionally. Best to explain using an
example. Let's say you have 2 arrays created from a time-dependent
procedure like this::

    a = []; b = []
    for t in range(100):
        if t % 3 == 0: a.append(t)
        if t % 5 == 0: b.append(1 - t)
    len(a), len(b) # returns (34, 20)

a and b are of different lengths, but you want to plot both from 30%
mark to 50% mark (for a, it's elements 10 -> 17, for b it's 6 -> 10),
as they are time-dependent. As you can probably tell, to get the indicies
10, 17, 6, 10 is messy. So, you can do something like this instead::

    r1, r2 = Range.proportionalSlice(Range(len(a)), Range(len(b)), slice(10, 17))

This will return the Ranges [10, 17] and [5.88, 10]

Then, you can plot both of them side by side like this::

    fig, axes = plt.subplots(ncols=2)
    axes[0].plot(r1.range_, a[r1.slice_])
    axes[1].plot(r2.range_, a[r2.slice_])
"""                                                                              # Range
        r1, r2 = Range(r1), Range(r2)                                            # Range
        ar1 = r1[r1Slice]; ar2 = r1.toRange(r2, ar1)                             # Range
        return ar1, ar2                                                          # Range
    def bound(self, rs:Union[range, slice]) -> Union[range, slice]:              # Range
        """If input range|slice's stop and start is missing, then use this
range's start and stop instead."""                                               # Range
        start = rs.start or self.start                                           # Range
        stop = rs.stop or self.stop                                              # Range
        return type(rs)(start, stop)                                             # Range
    def copy(self): return Range(self.start, self.stop)                          # Range
    def __str__(self): return f"[{self.start}, {self.stop}]"                     # Range
    def __eq__(self, _range):                                                    # Range
        _range = Range(_range)                                                   # Range
        return (_range.start == self.start or abs(_range.start - self.start) < 1e-9) and\
            (_range.stop == self.stop or abs(_range.stop - self.stop) < 1e-9)    # Range
    def __contains__(self, x:float): return x >= self.start and x < self.stop    # Range
    def __neg__(self): return Range(-self.start, -self.stop)                     # Range
    def __invert__(self): return Range(self.stop, self.start)                    # Range
    def __add__(self, num): return Range(self.start + num, self.stop + num)      # Range
    def __radd__(self, num): return self + num                                   # Range
    def __mul__(self, num): return Range(self.start * num, self.stop * num)      # Range
    def __rmul__(self, num): return self * num                                   # Range
    def __truediv__(self, num): return num * (1/num)                             # Range
    def __rtruediv__(self, num): raise "Doesn't make sense to do this!"          # Range
    def __round__(self): return Range(round(self.start), round(self.stop))       # Range
    def __ceil__(self): return Range(math.ceil(self.start), math.ceil(self.stop)) # Range
    def __floor__(self): return Range(math.floor(self.start), math.floor(self.stop)) # Range
    def __repr__(self):                                                          # Range
        return f"""A range of numbers: [{self.start}, {self.stop}]. Can do:
- r.toUnit(x): will convert x from range [{self.start}, {self.stop}] to [0, 1]
- r.fromUnit(x): will convert x from range [0, 1] to range [{self.start}, {self.stop}]
- r.toRange([a, b], x): will convert x from range [{self.start}, {self.stop}] to range [a, b]
- r[0], r[1], r.start, r.stop: get start and stop values of range

Note: for conversion methods, you can pass in"""                                 # Range
def yieldLowest(r1s:Iterator[Range], r2s:Iterator[Range]):                       # yieldLowest
    """Given 2 :class:`Range` generators with lengths a and b, yield every
object (a + b) so that :class:`Range`s with smaller start point gets yielded
first. Assumes that each generator:

- Does not intersect with itself
- Is sorted by start point already

.. warning::

    This method will sometimes yield the same objects given by the Iterators.
    Make sure you copy each :class:`Range` if your use case requires"""          # yieldLowest
    r1s = iter(r1s); r2s = iter(r2s)                                             # yieldLowest
    r1 = next(r1s, None)                                                         # yieldLowest
    if r1 is None: yield from r2s; return                                        # yieldLowest
    r2 = next(r2s, None)                                                         # yieldLowest
    if r2 is None: yield r1; yield from r1s; return                              # yieldLowest
    while True:                                                                  # yieldLowest
        while r1.start <= r2.start:                                              # yieldLowest
            yield r1                                                             # yieldLowest
            r1 = next(r1s, None)                                                 # yieldLowest
            if r1 is None: yield r2; yield from r2s; return                      # yieldLowest
        while r2.start <= r1.start:                                              # yieldLowest
            yield r2                                                             # yieldLowest
            r2 = next(r2s, None)                                                 # yieldLowest
            if r2 is None: yield r1; yield from r1s; return                      # yieldLowest
def join(r1s:Iterator[Range], r2s:Iterator[Range]):                              # join
    """Joins 2 :class:`Range` generators, so that overlaps gets merged
together.

.. warning::

    This method will sometimes yield the same objects given by the Iterators.
    Make sure you copy each :class:`Range` if your use case requires"""          # join
    it = yieldLowest(r1s, r2s); r = next(it, None)                               # join
    if r is None: return                                                         # join
    while True:                                                                  # join
        nr = next(it, None)                                                      # join
        if nr is None: yield r; return                                           # join
        if r.stop >= nr.start:                                                   # join
            r = r.copy(); r.stop = max(r.stop, nr.stop)                          # join
        else: yield r; r = nr                                                    # join
def intersect(r1s:Iterator[Range], r2s:Iterator[Range]):                         # intersect
    """Intersects 2 :class:`Range` generators, so that it only
returns overlaping regions"""                                                    # intersect
    r1s = iter(r1s); r2s = iter(r2s)                                             # intersect
    r1 = next(r1s, None)                                                         # intersect
    if r1 is None: return                                                        # intersect
    r2 = next(r2s, None)                                                         # intersect
    if r2 is None: return                                                        # intersect
    while True:                                                                  # intersect
        if True: # doesn't intersect at all                                      # intersect
            a = max(r1.start, r2.start)                                          # intersect
            b = min(r1.stop, r2.stop)                                            # intersect
            if a < b: yield Range(a, b)                                          # intersect
            if r1.stop > r2.stop: # loads next r2                                # intersect
                r2 = next(r2s, None)                                             # intersect
                if r2 is None: return                                            # intersect
            else: # loads next r1                                                # intersect
                r1 = next(r1s, None)                                             # intersect
                if r1 is None: return                                            # intersect
def neg(rs:List[Range]):                                                         # neg
    """Returns R - rs, where R is the set of real numbers."""                    # neg
    rs = iter(rs); r = next(rs, None)                                            # neg
    if r is None: yield Range(ninf, inf); return                                 # neg
    if ninf < r.start: yield Range(ninf, r.start) # check -inf case              # neg
    while True:                                                                  # neg
        start = r.stop                                                           # neg
        r = next(rs, None)                                                       # neg
        if r is None:                                                            # neg
            if start < inf: yield Range(start, inf)                              # neg
            return                                                               # neg
        yield Range(start, r.start)                                              # neg
class Domain:                                                                    # Domain
    def __init__(self, *ranges, dontCheck:bool=False):                           # Domain
        """Creates a new domain.

:param ranges: each element is a :class:`Range`, although any format will be fine as this selects for that
:param dontCheck: don't sanitize inputs, intended to boost perf internally only

A domain is just an array of :class:`Range` that represents what intervals on
the real number line is chosen. Some examples::

    inf = float("inf") # shorthand for infinity
    Domain([5, 7.5], [2, 3]) # represents "[2, 3) U [5, 7.5)"
    Domain([2, 3.2], [3, 8]) # represents "[2, 8)" as overlaps are merged
    -Domain([2, 3]) # represents "(-inf, 2) U [3, inf)", so essentially R - d, with R being the set of real numbers
    -Domain([-inf, 3]) # represents "[3, inf)"
    Domain.fromInts(2, 3, 6) # represents "[2, 4) U [6, 7)"

You can also do arithmetic on them, and check "in" oeprator::

    Domain([2, 3]) + Domain([4, 5]) # represents "[2, 3) U [4, 5)"
    Domain([2, 3]) + Domain([2.9, 5]) # represents "[2, 5)", also merges overlaps
    Domain([2, 3]) & Domain([2.5, 5]) # represents "[2, 3) A [2.5, 5)", or "[2.5, 3)"
    3 in Domain([2, 3]) # returns False
    2 in Domain([2, 3]) # returns True"""                                        # Domain
        if dontCheck: self.ranges = list(ranges); return                         # Domain
        # convert all to Range type, fix its order, and sort based on .start     # Domain
        ranges = [(r if isinstance(r, Range) else Range(r)).fixOrder() for r in ranges] # Domain
        ranges = sorted(ranges, key=lambda r: r.start)                           # Domain
        # merges overlapping segments                                            # Domain
        self.ranges = list(join(ranges, []))                                     # Domain
    @staticmethod                                                                # Domain
    def fromInts(*ints:List[int]):                                               # Domain
        """Returns a new :class:`Domain` which has ranges [i, i+1] for each
int given."""                                                                    # Domain
        return Domain(*(Range(i, i+1) for i in ints))                            # Domain
    def copy(self): return Domain(*(r.copy() for r in self.ranges))              # Domain
    def intIter(self, step:int=1, start:int=0):                                  # Domain
        """Yields ints in all ranges of this domain. If first range's domain
is :math:`(-\inf, a)`, then starts at the specified integer"""                   # Domain
        if len(self.ranges) == 0: return                                         # Domain
        for r in self.ranges:                                                    # Domain
            x = int(start) if r.start == -inf else int(r.start)                  # Domain
            while x < r.stop: yield x; x += step                                 # Domain
    def __neg__(self): return Domain(*neg(self.ranges), dontCheck=True)          # Domain
    def __add__(self, domain): return Domain(*(r.copy() for r in join(self.ranges, domain.ranges)), dontCheck=True) # Domain
    def __sub__(self, domain): return self + (-domain)                           # Domain
    def __and__(self, domain): return Domain(*intersect(self.ranges, domain.ranges), dontCheck=True) # Domain
    def __eq__(self, domain): return self.ranges == domain.ranges                # Domain
    def __str__(self): return f"Domain: {', '.join(str(r) for r in self.ranges)}" # Domain
    def __contains__(self, x): return any(x in r for r in self.ranges)           # Domain
    def __repr__(self):                                                          # Domain
        rs = '\n'.join(f"- {r}" for r in self.ranges)                            # Domain
        return f"""Domain:\n{rs}\n\nCan:
- 3 in d: check whether a number is in this domain or not
- d1 + d2: joins 2 domains
- -d: excludes the domain from R
- d1 - d2: same as d1 + (-d2)
- d1 & d2: intersects 2 domains"""                                               # Domain
puas = [[ord(c) for c in cs] for cs in [["\ue000", "\uf8ff"], ["\U000f0000", "\U000ffffd"], ["\U00100000", "\U0010fffd"]]] # Domain
class AutoIncrement:                                                             # AutoIncrement
    def __init__(self, initialValue:int=-1, n:int=float("inf"), prefix:str=None): # AutoIncrement
        """Creates a new AutoIncrement object. Every time the object is called
it gets incremented by 1 automatically. Example::

    a = k1lib.AutoIncrement()
    a() # returns 0
    a() # returns 1
    a() # returns 2
    a.value # returns 2
    a.value # returns 2
    a() # returns 3

    a = AutoIncrement(n=3, prefix="cluster_")
    a() # returns "cluster_0"
    a() # returns "cluster_1"
    a() # returns "cluster_2"
    a() # returns "cluster_0"

:param n: if specified, then will wrap around to 0 when hit this number
:param prefix: if specified, will yield strings with specified prefix"""         # AutoIncrement
        self.value = initialValue; self.n = n; self.prefix = prefix              # AutoIncrement
    @staticmethod                                                                # AutoIncrement
    def random() -> "AutoIncrement":                                             # AutoIncrement
        """Creates a new AutoIncrement object that has a random integer initial value""" # AutoIncrement
        return AutoIncrement(random.randint(0, 1e9))                             # AutoIncrement
    @property                                                                    # AutoIncrement
    def value(self):                                                             # AutoIncrement
        """Get the value as-is, without auto incrementing it"""                  # AutoIncrement
        if self.prefix is None: return self._value                               # AutoIncrement
        return f"{self.prefix}{self._value}"                                     # AutoIncrement
    @value.setter                                                                # AutoIncrement
    def value(self, value): self._value = value                                  # AutoIncrement
    def __call__(self):                                                          # AutoIncrement
        """Increments internal counter, and return it."""                        # AutoIncrement
        self._value += 1                                                         # AutoIncrement
        if self._value >= self.n: self._value = 0                                # AutoIncrement
        return self.value                                                        # AutoIncrement
    @staticmethod                                                                # AutoIncrement
    def unicode_pua():                                                           # AutoIncrement
        """Returns a generator that generates unicode characters from within
unicode's private use area (PUA). Example::

    a = k1.AutoIncrement.unicode_pua()
    a | head() | deref() # returns ['\ue000', '\ue001', '\ue002', '\ue003', '\ue004', '\ue005', '\ue006', '\ue007', '\ue008', '\ue009']
"""                                                                              # AutoIncrement
        for pua in puas:                                                         # AutoIncrement
            for c in range(pua[0], pua[1]+1): yield chr(c)                       # AutoIncrement
class Wrapper:                                                                   # Wrapper
    value:Any                                                                    # Wrapper
    """Internal value of this :class:`Wrapper`"""                                # Wrapper
    def __init__(self, value=None):                                              # Wrapper
        """Creates a wrapper for some value and get it by calling it.
Example::

    a = k1.Wrapper(list(range(int(1e7))))
    # returns [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    a()[:10]

This exists just so that Jupyter Lab's contextual help won't automatically
display the (possibly humongous) value. Could be useful if you want to pass a
value by reference everywhere like this::

    o = k1.Wrapper(None)
    def f(obj):
        obj.value = 3
    f(o)
    o() # returns 3

You can also pipe into it like this:

    o = 3 | k1.Wrapper()
    o() # returns 3
"""                                                                              # Wrapper
        self.value = value                                                       # Wrapper
    def __call__(self): return self.value                                        # Wrapper
    def __ror__(self, it): self.value = it; return self                          # Wrapper
class Every:                                                                     # Every
    def __init__(self, n):                                                       # Every
        """Returns True every interval.
Example::

    e = k1lib.Every(4)
    e() # returns True
    e() # returns False
    e() # returns False
    e() # returns False
    e() # returns True"""                                                        # Every
        self.n = n; self.i = -1                                                  # Every
    def __call__(self) -> bool:                                                  # Every
        """Returns True or False based on internal count."""                     # Every
        self.i += 1; return self.value                                           # Every
    @property                                                                    # Every
    def value(self) -> bool:                                                     # Every
        if self.i % self.n: return False                                         # Every
        else: return True                                                        # Every
class RunOnce:                                                                   # RunOnce
    def __init__(self):                                                          # RunOnce
        """Returns False first time only.
Example::

    r = k1lib.RunOnce()
    r.done() # returns False
    r.done() # returns True
    r.done() # returns True
    r.revert()
    r.done() # returns False
    r.done() # returns True
    r.done() # returns True

May be useful in situations like::

    class A:
        def __init__(self):
            self.ro = k1lib.RunOnce()
        def f(self, x):
            if self.ro.done(): return 3 + x
            return 5 + x
    a = A()
    a.f(4) # returns 9
    a.f(4) # returns 7"""                                                        # RunOnce
        self.value = False                                                       # RunOnce
    def done(self):                                                              # RunOnce
        """Whether this has been called once before."""                          # RunOnce
        v = self.value                                                           # RunOnce
        self.value = True                                                        # RunOnce
        return v                                                                 # RunOnce
    def __call__(self):                                                          # RunOnce
        """Alias of :meth:`done`."""                                             # RunOnce
        return self.done()                                                       # RunOnce
    def revert(self):                                                            # RunOnce
        self.value = False                                                       # RunOnce
class MaxDepth:                                                                  # MaxDepth
    def __init__(self, maxDepth:int, depth:int=0):                               # MaxDepth
        """Convenience utility to check for graph max depth.
Example::

    def f(d):
        print(d.depth)
        if d: f(d.enter())
    # prints "0\\n1\\n2\\n3"
    f(k1lib.MaxDepth(3))

Of course, this might look unpleasant to the end user, so this is more
likely for internal tools."""                                                    # MaxDepth
        self.maxDepth = maxDepth; self.depth = depth                             # MaxDepth
    def enter(self) -> "MaxDepth":                                               # MaxDepth
        return MaxDepth(self.maxDepth, self.depth + 1)                           # MaxDepth
    def __bool__(self):                                                          # MaxDepth
        return self.depth < self.maxDepth                                        # MaxDepth
    def __call__(self):                                                          # MaxDepth
        """Alias of :meth:`__bool__`."""                                         # MaxDepth
        return bool(self)                                                        # MaxDepth
class MovingAvg:                                                                 # MovingAvg
    def __init__(self, initV:float=0, alpha=0.9, debias=False):                  # MovingAvg
        """Smoothes out sequential data using momentum.
Example::

    a = k1lib.MovingAvg(5)
    a(3).value # returns 4.8, because 0.9*5 + 0.1*3 = 4.8
    a(3).value # returns 4.62

There's also a cli at :class:`~k1lib.cli.conv.toMovingAvg` that does the exact
same thing, but just more streamlined and cli-like. Both versions are kept as
sometimes I do want a separate object with internal state

Difference between normal and debias modes::

    x = torch.linspace(0, 10, 100); y = torch.cos(x) | op().item().all() | deref()
    plt.plot(x, y);
    a = k1lib.MovingAvg(debias=False); plt.plot(x, y | apply(lambda y: a(y).value) | deref())
    a = k1lib.MovingAvg(debias=True); plt.plot(x, y | apply(lambda y: a(y).value) | deref())
    plt.legend(["Signal", "Normal", "Debiased"])

.. image:: images/movingAvg.png

As you can see, normal mode still has the influence of the initial value at
0 and can't rise up fast, whereas the debias mode will ignore the initial
value and immediately snaps to the first saved value.

:param initV: initial value
:param alpha: number in [0, 1]. Basically how much to keep old value?
:param debias: whether to debias the initial value"""                            # MovingAvg
        self.value = initV; self.alpha = alpha; self.debias = debias             # MovingAvg
        self.m = self.value; self.t = 0                                          # MovingAvg
    def __call__(self, value):                                                   # MovingAvg
        """Updates the average with a new value"""                               # MovingAvg
        self.m = self.m * self.alpha + value * (1 - self.alpha)                  # MovingAvg
        if self.debias:                                                          # MovingAvg
            self.t += 1                                                          # MovingAvg
            self.value = self.m / (1 - self.alpha**self.t)                       # MovingAvg
        else: self.value = self.m                                                # MovingAvg
        return self                                                              # MovingAvg
    def __add__(self, o): return self.value + o                                  # MovingAvg
    def __radd__(self, o): return o + self.value                                 # MovingAvg
    def __sub__(self, o): return self.value - o                                  # MovingAvg
    def __rsub__(self, o): return o - self.value                                 # MovingAvg
    def __mul__(self, o): return self.value * o                                  # MovingAvg
    def __rmul__(self, o): return o * self.value                                 # MovingAvg
    def __truediv__(self, o): return self.value / o                              # MovingAvg
    def __rtruediv__(self, o): return o / self.value                             # MovingAvg
    def __repr__(self):                                                          # MovingAvg
        return f"Moving average: {self.value}, alpha: {self.alpha}"              # MovingAvg
sen = "_ab_sentinel"                                                             # MovingAvg
jitOpcodes = {"__len__": lambda x: f"len({x})",                                  # MovingAvg
              "__neg__": lambda x: f"(-{x})",                                    # MovingAvg
              "__pos__": lambda x: f"(+{x})",                                    # MovingAvg
              "__abs__": lambda x: f"abs({x})",                                  # MovingAvg
              "__invert__": lambda x: f"(~{x})",                                 # MovingAvg
              "__getattr__": lambda x, idx: f"getattr({x},{idx})",               # MovingAvg
              "__getitem__": lambda x, idx: f"({x}[{idx}])",                     # MovingAvg
              "__round__": lambda x, o: f"round({x}, {o})",                      # MovingAvg
              "__add__": lambda x, o: f"({x}+{o})",                              # MovingAvg
              "__radd__": lambda x, o: f"({o}+{x})",                             # MovingAvg
              "__sub__": lambda x, o: f"({x}-{o})",                              # MovingAvg
              "__rsub__": lambda x, o: f"({o}-{x})",                             # MovingAvg
              "__mul__": lambda x, o: f"({x}*{o})",                              # MovingAvg
              "__rmul__": lambda x, o: f"({o}*{x})",                             # MovingAvg
              "__matmul__": lambda x, o: f"({x}@{o})",                           # MovingAvg
              "__rmatmul__": lambda x, o: f"({o}@{x})",                          # MovingAvg
              "__truediv__": lambda x, o: f"({x}/{o})",                          # MovingAvg
              "__rtruediv__": lambda x, o: f"({o}/{x})",                         # MovingAvg
              "__floordiv__": lambda x, o: f"({x}//{o})",                        # MovingAvg
              "__rfloordiv__": lambda x, o: f"({o}//{x})",                       # MovingAvg
              "__mod__": lambda x, o: f"({x}%{o})",                              # MovingAvg
              "__rmod__": lambda x, o: f"({o}%{x})",                             # MovingAvg
              "__pow__": lambda x, o: f"({x}**{o})",                             # MovingAvg
              "__rpow__": lambda x, o: f"({o}**{x})",                            # MovingAvg
              "__lshift__": lambda x, o: f"({x}<<{o})",                          # MovingAvg
              "__rlshift__": lambda x, o: f"({o}<<{x})",                         # MovingAvg
              "__rshift__": lambda x, o: f"({x}>>{o})",                          # MovingAvg
              "__rrshift__": lambda x, o: f"({o}>>{x})",                         # MovingAvg
              "__and__": lambda x, o: f"({x}&{o})",                              # MovingAvg
              "__rand__": lambda x, o: f"({o}&{x})",                             # MovingAvg
              "__xor__": lambda x, o: f"({x}^{o})",                              # MovingAvg
              "__rxor__": lambda x, o: f"({o}^{x})",                             # MovingAvg
              "__or__": lambda x, o: f"({x}|{o})",                               # MovingAvg
              "__ror__": lambda x, o: f"({o}|{x})",                              # MovingAvg
              "__lt__": lambda x, o: f"({x}<{o})",                               # MovingAvg
              "__le__": lambda x, o: f"({x}<={o})",                              # MovingAvg
              "__eq__": lambda x, o: f"({x}=={o})",                              # MovingAvg
              "__ne__": lambda x, o: f"({x}!={o})",                              # MovingAvg
              "__gt__": lambda x, o: f"({x}>{o})",                               # MovingAvg
              "__ge__": lambda x, o: f"({x}>={o})",}                             # MovingAvg
opcodeAuto = AutoIncrement(prefix=f"_op_{random.randint(100,999)}_var_")         # MovingAvg
compareOps = {"__lt__", "__le__", "__eq__", "__ne__", "__gt__", "__ge__"}        # MovingAvg
class Absorber:                                                                  # Absorber
    """Creates an object that absorbes every operation done on it. Could be
useful in some scenarios::

    ab = k1lib.Absorber()
    # absorbs all operations done on the object
    abs(ab[::3].sum(dim=1))

    t = torch.randn(5, 3, 3)
    # returns transformed tensor of size [2, 3]
    ab.ab_operate(t)

Another::

    ab = Absorber()
    ab[2] = -50
    # returns [0, 1, -50, 3, 4]
    ab.ab_operate(list(range(5)))

Because this object absorbs every operation done on it, you have to be gentle with
it, as any unplanned disturbances might throw your code off. Best to create a new
one on the fly, and pass them immediately to functions, because if you're in a
notebook environment like Jupyter, it might poke at variables.

For extended code example that utilizes this, check over :class:`k1lib.cli.modifier.op`
source code."""                                                                  # Absorber
    def __init__(self, initDict:dict=dict()):                                    # Absorber
        """Creates a new Absorber.

:param initDict: initial variables to set, as setattr operation is normally absorbed""" # Absorber
        self._ab_sentinel = True                                                 # Absorber
        self._ab_steps = []                                                      # Absorber
        self._ab_solidified = False                                              # Absorber
        for k, v in initDict.items(): setattr(self, k, v)                        # Absorber
        self._ab_sentinel = False                                                # Absorber
    def ab_solidify(self):                                                       # Absorber
        """Use this to not absorb ``__call__`` operations anymore and makes it
feel like a regular function (still absorbs other operations though)::

    f = op()**2
    3 | f # returns 9, but may be you don't want to pipe it in
    f.op_solidify()
    f(3)  # returns 9"""                                                         # Absorber
        self._ab_sentinel = True                                                 # Absorber
        self._ab_solidified = True                                               # Absorber
        self._ab_sentinel = False                                                # Absorber
        return self                                                              # Absorber
    def ab_operate(self, x):                                                     # Absorber
        """Special method to actually operate on an object and get the result. Not
absorbed. Example::

    # returns 6
    (op() * 2).ab_operate(3)"""                                                  # Absorber
        for desc, step in self._ab_steps: x = step(x)                            # Absorber
        return x                                                                 # Absorber
    def ab_fastFS(self) -> str:                                                  # Absorber
        s = self._ab_steps; l = len(s)                                           # Absorber
        x = k1lib.cli.init._jsDAuto()                                            # Absorber
        try: # jit compilation                                                   # Absorber
            ss = x; values = {}                                                  # Absorber
            for (opcode, *o), *_ in s:                                           # Absorber
                if opcode == "__call__":                                         # Absorber
                    va = opcodeAuto(); vk = opcodeAuto()                         # Absorber
                    values[va], values[vk] = o[0]                                # Absorber
                    ss = f"({ss}(*{va}, **{vk}))"                                # Absorber
                elif len(o) > 0:                                                 # Absorber
                    varname = opcodeAuto(); v = o[0]                             # Absorber
                    if isinstance(v, (int, float)):                              # Absorber
                        ss = jitOpcodes[opcode](ss, v)                           # Absorber
                    else:                                                        # Absorber
                        values[varname] = v                                      # Absorber
                        ss = jitOpcodes[opcode](ss, varname)                     # Absorber
                else: ss = jitOpcodes[opcode](ss)                                # Absorber
            return [f"lambda {x}: {ss}", values]                                 # Absorber
        except Exception as e: pass                                              # Absorber
    def ab_fastF(self):                                                          # Absorber
        """Returns a function that operates on the input (just like :meth:`ab_operate`),
but much faster, suitable for high performance tasks. Example::

    f = (k1lib.Absorber() * 2).ab_fastF()
    # returns 6
    f(3)"""                                                                      # Absorber
        s = self._ab_steps; l = len(s)                                           # Absorber
        res = self.ab_fastFS() # jit compilation, compressing multiple steps to a single simple expression # Absorber
        if res: fn, values = res; return eval(compile(fn, "", "eval"), values)   # Absorber
        if l == 0: return lambda x: x                                            # Absorber
        if l == 1: return s[0][1]                                                # Absorber
        if l == 2:                                                               # Absorber
            a, b = s[0][1], s[1][1]                                              # Absorber
            return lambda x: b(a(x))                                             # Absorber
        if l == 3:                                                               # Absorber
            a, b, c = s[0][1], s[1][1], s[2][1]                                  # Absorber
            return lambda x: c(b(a(x)))                                          # Absorber
        if l == 4:                                                               # Absorber
            a, b, c, d = s[0][1], s[1][1], s[2][1], s[3][1]                      # Absorber
            return lambda x: d(c(b(a(x))))                                       # Absorber
        if l == 5:                                                               # Absorber
            a, b, c, d, e = s[0][1], s[1][1], s[2][1], s[3][1], s[4][1]          # Absorber
            return lambda x: e(d(c(b(a(x)))))                                    # Absorber
        return self.ab_operate                                                   # Absorber
    def _ab_steps_append(self, o):                                               # Absorber
        if not self._ab_solidified:                                              # Absorber
            s = self._ab_steps                                                   # Absorber
            if len(s) > 0 and s[-1][0][0] in compareOps and o[0][0] in compareOps: # Absorber
                p = s.pop(); a = p[1]; b = o[1] # for 2 consecutive compare operations # Absorber
                s.append([["compareOps"], lambda x: a(x) and b(x)])              # Absorber
            else: s.append(o)                                                    # Absorber
        return self                                                              # Absorber
    def __getattr__(self, idx):                                                  # Absorber
        if isinstance(idx, str) and idx.startswith("_"): raise AttributeError("Getting attributes starting with underscore is prohibited. If you're using `op`, consider using `aS(lambda x: x._field)` instead.") # Absorber
        return self._ab_steps_append([["__getattr__", idx], lambda x: getattr(x, idx)]); # Absorber
    def __setattr__(self, k, v):                                                 # Absorber
        """Only allows legit variable setting when '_ab_sentinel' is True. Absorbs
operations if it's False."""                                                     # Absorber
        if k == sen: self.__dict__[k] = v                                        # Absorber
        else:                                                                    # Absorber
            if self.__dict__[sen]: self.__dict__[k] = v                          # Absorber
            else:                                                                # Absorber
                def f(x): setattr(x, k, v); return x                             # Absorber
                self._ab_steps_append([["__setattr__", [k, v]], f])              # Absorber
                if self._ab_solidified: self.__dict__[k] = v                     # Absorber
                return self                                                      # Absorber
    def __getitem__(self, idx): return self._ab_steps_append([["__getitem__", idx], lambda x: x[idx]]); # Absorber
    def __setitem__(self, k, v):                                                 # Absorber
        def f(x): x[k] = v; return x                                             # Absorber
        return self._ab_steps_append([["__setitem__", [k, v]], f]);              # Absorber
    def __call__(self, *args, **kwargs): return self._ab_steps_append([["__call__", [args, kwargs]], lambda x: x(*args, **kwargs)]); # Absorber
    def __round__(self, ndigits=0):   return self._ab_steps_append([["__round__",       ndigits], lambda x: round(x, ndigits)]); # Absorber
    def __add__(self, o):       return self._ab_steps_append([["__add__",       o], lambda x: x+o ]); # Absorber
    def __radd__(self, o):      return self._ab_steps_append([["__radd__",      o], lambda x: o+x ]); # Absorber
    def __sub__(self, o):       return self._ab_steps_append([["__sub__",       o], lambda x: x-o ]); # Absorber
    def __rsub__(self, o):      return self._ab_steps_append([["__rsub__",      o], lambda x: o-x ]); # Absorber
    def __mul__(self, o):       return self._ab_steps_append([["__mul__",       o], lambda x: x*o ]); # Absorber
    def __rmul__(self, o):      return self._ab_steps_append([["__rmul__",      o], lambda x: o*x ]); # Absorber
    def __matmul__(self, o):    return self._ab_steps_append([["__matmul__",    o], lambda x: x@o ]); # Absorber
    def __rmatmul__(self, o):   return self._ab_steps_append([["__rmatmul__",   o], lambda x: o@x ]); # Absorber
    def __truediv__(self, o):   return self._ab_steps_append([["__truediv__",   o], lambda x: x/o ]); # Absorber
    def __rtruediv__(self, o):  return self._ab_steps_append([["__rtruediv__",  o], lambda x: o/x ]); # Absorber
    def __floordiv__(self, o):  return self._ab_steps_append([["__floordiv__",  o], lambda x: x//o]); # Absorber
    def __rfloordiv__(self, o): return self._ab_steps_append([["__rfloordiv__", o], lambda x: o//x]); # Absorber
    def __mod__(self, o):       return self._ab_steps_append([["__mod__",       o], lambda x: x%o ]); # Absorber
    def __rmod__(self, o):      return self._ab_steps_append([["__rmod__",      o], lambda x: o%x ]); # Absorber
    def __pow__(self, o):       return self._ab_steps_append([["__pow__",       o], lambda x: x**o]); # Absorber
    def __rpow__(self, o):      return self._ab_steps_append([["__rpow__",      o], lambda x: o**x]); # Absorber
    def __lshift__(self, o):    return self._ab_steps_append([["__lshift__",    o], lambda x: x<<o]); # Absorber
    def __rlshift__(self, o):   return self._ab_steps_append([["__rlshift__",   o], lambda x: o<<x]); # Absorber
    def __rshift__(self, o):    return self._ab_steps_append([["__rshift__",    o], lambda x: x>>o]); # Absorber
    def __rrshift__(self, o):   return self._ab_steps_append([["__rrshift__",   o], lambda x: o>>x]); # Absorber
    def __and__(self, o):       return self._ab_steps_append([["__and__",       o], lambda x: x&o ]); # Absorber
    def __rand__(self, o):      return self._ab_steps_append([["__rand__",      o], lambda x: o&x ]); # Absorber
    def __xor__(self, o):       return self._ab_steps_append([["__xor__",       o], lambda x: x^o ]); # Absorber
    def __rxor__(self, o):      return self._ab_steps_append([["__rxor__",      o], lambda x: o^x ]); # Absorber
    def __or__(self, o):        return self._ab_steps_append([["__or__",        o], lambda x: x|o ]); # Absorber
    def __ror__(self, o):       return self._ab_steps_append([["__ror__",       o], lambda x: o|x ]); # Absorber
    def __lt__(self, o):        return self._ab_steps_append([["__lt__",        o], lambda x: x<o ]); # Absorber
    def __le__(self, o):        return self._ab_steps_append([["__le__",        o], lambda x: x<=o]); # Absorber
    def __eq__(self, o):        return self._ab_steps_append([["__eq__",        o], lambda x: x==o]); # Absorber
    def __ne__(self, o):        return self._ab_steps_append([["__ne__",        o], lambda x: x!=o]); # Absorber
    def __gt__(self, o):        return self._ab_steps_append([["__gt__",        o], lambda x: x>o ]); # Absorber
    def __ge__(self, o):        return self._ab_steps_append([["__ge__",        o], lambda x: x>=o]); # Absorber
    def __neg__(self):    return self._ab_steps_append([["__neg__"],    lambda x: -x      ]); # Absorber
    def __pos__(self):    return self._ab_steps_append([["__pos__"],    lambda x: +x      ]); # Absorber
    def __abs__(self):    return self._ab_steps_append([["__abs__"],    lambda x: abs(x)  ]); # Absorber
    def __invert__(self): return self._ab_steps_append([["__invert__"], lambda x: ~x      ]); # Absorber
    def ab_int(self):                                                            # Absorber
        """Replacement for ``int(ab)``, as that requires returning an actual :class:`int`.""" # Absorber
        return self._ab_steps_append([["__int__"],    lambda x: int(x)  ]);      # Absorber
    def __int__(self):    return self.int()                                      # Absorber
    def ab_float(self):                                                          # Absorber
        """Replacement for ``float(ab)``, as that requires returning an actual :class:`float`.""" # Absorber
        return self._ab_steps_append([["__float__"],  lambda x: float(x)]);      # Absorber
    def __float__(self):  return self.float()                                    # Absorber
    def ab_str(self):                                                            # Absorber
        """Replacement for ``str(ab)``, as that requires returning an actual :class:`str`.""" # Absorber
        return self._ab_steps_append([["__str__"],    lambda x: str(x)  ]);      # Absorber
    def ab_len(self):                                                            # Absorber
        """Replacement for ``len(ab)``, as that requires returning an actual :class:`int`.""" # Absorber
        return self._ab_steps_append([["__len__"],    lambda x: len(x)  ]);      # Absorber
    def ab_contains(self, key):                                                  # Absorber
        """Replacement for ``key in ab``, as that requires returning an actual :class:`int`.""" # Absorber
        return self._ab_steps_append([["__contains__", key], lambda x: key in x]); # Absorber
sep = "\u200b" # weird separator, guaranteed (mostly) to not appear anywhere in the # Absorber
# settings, so that I can pretty print it                                        # Absorber
class Settings:                                                                  # Settings
    def __init__(self, **kwargs):                                                # Settings
        """Creates a new settings object. Basically fancy version of :class:`dict`.
Example::

    s = k1lib.Settings(a=3, b="42")
    s.c = k1lib.Settings(d=8)

    s.a # returns 3
    s.b # returns "42"
    s.c.d # returns 8
    print(s) # prints nested settings nicely"""                                  # Settings
        self._setattr_sentinel = True                                            # Settings
        for k, v in kwargs.items(): setattr(self, k, v)                          # Settings
        self._docs = dict(); self._cbs = dict()                                  # Settings
        self._setattr_sentinel = False                                           # Settings
    @contextlib.contextmanager                                                   # Settings
    def context(self, **kwargs):                                                 # Settings
        """Context manager to temporarily modify some settings. Applies
to all sub-settings. Example::

    s = k1lib.Settings(a=3, b="42", c=k1lib.Settings(d=8))
    with s.context(a=4):
        s.c.d = 20
        s.a # returns 4
        s.c.d # returns 20
    s.a # returns 3
    s.c.d # returns 8"""                                                         # Settings
        oldValues = dict(self.__dict__); err = None                              # Settings
        for k in kwargs.keys():                                                  # Settings
            if k not in oldValues:                                               # Settings
                raise RuntimeError(f"'{k}' settings not found!")                 # Settings
        try:                                                                     # Settings
            with contextlib.ExitStack() as stack:                                # Settings
                for _, sub in self._subSettings():                               # Settings
                    stack.enter_context(sub.context())                           # Settings
                for k, v in kwargs.items(): setattr(self, k, v)                  # Settings
                yield                                                            # Settings
        finally:                                                                 # Settings
            for k, v in oldValues.items(): setattr(self, k, v)                   # Settings
    def add(self, k:str, v:Any, docs:str="", cb:Callable[["Settings", Any], None]=None) -> "Settings": # Settings
        """Long way to add a variable. Advantage of this is that you can slip in extra
documentation for the variable. Example::

    s = k1lib.Settings()
    s.add("a", 3, "some docs")
    print(s) # displays the extra docs

:param cb: callback that takes in (settings, new value) if any property changes""" # Settings
        setattr(self, k, v); self._docs[k] = docs                                # Settings
        self._cbs[k] = cb; return self                                           # Settings
    def _docsOf(self, k:str):                                                    # Settings
        return f"{self._docs[k]}" if k in self._docs else ""                     # Settings
    def _subSettings(self) -> List[Tuple[str, "Settings"]]:                      # Settings
        return [(k, v) for k, v in self.__dict__.items() if isinstance(v, Settings) and not k.startswith("_")] # Settings
    def _simpleSettings(self) -> List[Tuple[str, Any]]:                          # Settings
        return [(k, v) for k, v in self.__dict__.items() if not isinstance(v, Settings) and not k.startswith("_")] # Settings
    def __setattr__(self, k, v):                                                 # Settings
        self.__dict__[k] = v                                                     # Settings
        if k != "_setattr_sentinel" and not self._setattr_sentinel:              # Settings
            if k in self._cbs and self._cbs[k] is not None: self._cbs[k](self, v) # Settings
    def __repr__(self):                                                          # Settings
        """``includeDocs`` mainly used internally when generating docs in sphinx.""" # Settings
        ks = list(k for k in self.__dict__ if not k.startswith("_"))             # Settings
        kSpace = max([1, *(ks | k1lib.cli.shape(0).all())]); s = "Settings:\n"   # Settings
        for k, v in self._simpleSettings():                                      # Settings
            s += f"- {k.ljust(kSpace)} = {k1lib.limitChars(str(v), settings.displayCutoff)}{sep}{self._docsOf(k)}\n" # Settings
        for k, v in self._subSettings():                                         # Settings
            sub = v.__repr__().split("\n")[1:-1] | k1lib.cli.tab("  ") | k1lib.cli.join("\n") # Settings
            s += f"- {k.ljust(kSpace)} = <Settings>{sep}{self._docsOf(k)}\n" + sub + "\n" # Settings
        return s.split("\n") | k1lib.cli.op().split(sep).all() | k1lib.cli.pretty(sep) | k1lib.cli.join("\n") # Settings
_settings = Settings().add("test", Settings().add("bio", True, "whether to test bioinformatics clis that involve strange command line tools like samtools and bwa")) # Settings
_settings.add("packages", Settings(), "which package is available to use?")      # Settings
settings = Settings().add("displayCutoff", 50, "cutoff length when displaying a Settings object") # Settings
settings.add("svgScale", 0.7, "default svg scales for clis that displays graphviz graphs") # Settings
def _cb_wd(s, p):                                                                # _cb_wd
    if p != None: p = os.path.abspath(os.path.expanduser(p)); _oschdir(p)        # _cb_wd
    s.__dict__["wd"] = p                                                         # _cb_wd
def oschdir(path): settings.wd = path                                            # oschdir
_oschdir = os.chdir; os.chdir = oschdir; os.chdir.__doc__ = _oschdir.__doc__     # oschdir
settings.add("wd", os.getcwd(), "default working directory, will get from `os.getcwd()`. Will update using `os.chdir()` automatically when changed", _cb_wd) # oschdir
settings.add("cancelRun_newLine", True, "whether to add a new line character at the end of the cancel run/epoch/batch message") # oschdir
or_patch = Settings()\
    .add("numpy", True, "whether to patch numpy arrays")\
    .add("dict", True, "whether to patch Python dict keys and items")\
    .add("pandas", True, "whether to patch pandas series")                       # oschdir
startup = Settings().add("init_ray", True, "whether to connect to ray's cluster accessible locally automatically")\
    .add("import_optionals", True, "whether to try to import optional dependencies automatically or not. Set this to False if you want a faster load time, but with reduced functionalities")\
    .add("or_patch", or_patch, "whether to patch __or__() method for several C-extension datatypes (numpy array, pandas data frame/series, etc). This would make cli operations with them a lot more pleasant, but might cause strange bugs. Haven't met them myself though") # oschdir
settings.add("startup", startup, "these settings have to be applied like this: `import k1lib; k1lib.settings.startup.or_patch = False; from k1lib.imports import *` to ensure that the values are set") # oschdir
settings.add("pushNotificationKey", os.getenv("k1lib_pushNotificationKey", None), "API key for `k1lib.pushNotification()`. See docs of that for more info") # oschdir
def sign(v): return 1 if v > 0 else -1                                           # sign
def roundOff(a, b):                                                              # roundOff
    m = (a + b) / 2                                                              # roundOff
    return m                                                                     # roundOff
    dec = math.log10(abs(a-m)+1e-7) # decimal place                              # roundOff
    factor = 10**(sign(dec) * math.floor(abs(dec)+1e-7)+1)                       # roundOff
    return factor*round(m/factor)                                                # roundOff
def toPrecision(num, sig=1):                                                     # toPrecision
    try:                                                                         # toPrecision
        if num == 0: return 0                                                    # toPrecision
        s = sign(num); num = abs(num)                                            # toPrecision
        fac = 10**(-math.floor(math.log10(num))+sig-1)                           # toPrecision
        return s*round(num*fac)/fac                                              # toPrecision
    except: return num                                                           # toPrecision
def niceUS(mean, std):                                                           # niceUS
    try:                                                                         # niceUS
        if std < 1e-12: return mean, std                                         # niceUS
        pres = 2 if std/10**math.floor(math.log10(std)) < 2 else 1               # niceUS
        std2 = toPrecision(std, pres)                                            # niceUS
        fac = 10**(-math.floor(math.log10(std2))+pres-1)                         # niceUS
        return round(mean*fac)/fac, std2                                         # niceUS
    except: return mean, std                                                     # niceUS
def removeOutliers(t, fraction=0.01):                                            # removeOutliers
    b = int(len(t)*fraction/2)                                                   # removeOutliers
    return np.sort(t)[b:-b]                                                      # removeOutliers
def _US(v): return [*v] if isinstance(v, UValue) else [v, 0]                     # _US
if True: # used to be a hasTorch condition here, lazy to reindent                # _US
    @functools.total_ordering                                                    # _US
    class UValue:                                                                # _US
        _unit = np.random.randn(2, 5, 100000)                                    # _US
        def __init__(self, mean=0, std=1, N=None):                               # _US
            """Creates a new "uncertain value", which has a mean and a standard
deviation. You can then do math operations on them as normal, and the
propagation errors will be automatically calculated for you. Make sure to
run the calculation multiple times as the mean and std values fluctuates by
a little run-by-run. Example::

    # returns UValue(mean=4.7117, std=3.4736) object
    abs(k1lib.UValue() * 5 + 3)

You can also instantiate from an existing list/numpy array/pytorch tensor::

    # returns UValue(mean=24.5, std=14.431) object
    k1lib.UValue.fromSeries(range(50))

You can also do arbitrary complex math operations::

    # returns UValue(mean=0.5544, std=0.4871)
    (20 + k1lib.UValue()).f(np.sin)
    # same as above, but takes longer to run!
    (20 + k1lib.UValue()).f(math.sin)

I suggest you to make your arbitrary function out of numpy's operations,
as those are a fair bit faster than regular Python.

If you have a list of :class:`UValue`, and want to plot them with error
bars, then you can do something like this::

    x = np.linspace(0, 6)
    y = list(np.sin(x)*10) | apply(k1lib.UValue) | toList()
    plt.errorbar(x, *(y | transpose()));

There are several caveats however:

.. note::

    First is the problem of theoretically vs actually sample a
    distribution. Let's see an example::

        # returns theoretical value UValue(mean=8000.0, std=1200.0) -> 8000.0 ± 1200.0
        k1lib.UValue(20) ** 3
        # prints out actual mean and std value of (8064.1030, 1204.3529)
        a = k1lib.UValue(20).sample() ** 3
        print(a.mean(), a.std())

    So far so good. However, let's create some uncertainty in "3"::

        # returns theoretical value UValue(mean=8000.0, std=23996.0) -> 10000.0 ± 20000.0
        k1lib.UValue(20) ** k1lib.UValue(3)
        # prints out actual mean and std value of (815302.8750, 27068828.), but is very unstable and changes a lot
        a = k1lib.UValue(20).sample() ** k1lib.UValue(3).sample()
        print(a.mean(), a.std())

    Woah, what happens here? The actual mean and std values are
    completely different from the theoretical values. This is
    mainly due to UValue(3) has some outlier values large enough
    to boost the result up multiple times. Even removing 1% of
    values on either end of the spectrum does not quite work. So,
    becareful to interpret these uncertainty values, and in some
    case the theoretical estimates from math are actually very
    unstable and will not be observed in real life.

.. note::

    Then there's the problem of each complex operation, say ``(v*2+3)/5``
    will be done step by step, meaning ``a=v*2`` mean and std will be
    calculated first, then ignoring the calculated sample values and just
    go with the mean and std, sample a bunch of values from there and calculate
    ``a+3`` mean and std. Rinse and repeat. This means that these 2 statements
    may differ by a lot::

        # prints out (0.15867302766786406, 0.12413313456900205)
        x = np.linspace(-3, 3, 1000); sq = (abs(x)-0.5)**2; y = sq*np.exp(-sq)
        print(y.mean(), y.std())

        # returns UValue(mean=0.081577, std=0.32757) -> 0.1 ± 0.3
        x = k1lib.UValue(0, 1); sq = (abs(x)-0.5)**2; y = sq*(-sq).f(np.exp)

    Why this weird function? It converts from a single nice hump into multiple
    complex humps. Anyway, this serves to demonstrate that the result from the
    ``calculate -> get mean, std -> sample from new distribution -> calculate``
    process might be different from just calculating from start to end and then
    get the mean and std.

.. note::

    Lastly, you might have problems when using the same UValue multiple times in
    an expression::

        a = UValue(10, 1)
        a * 2 # has mean 20, std 2
        a + a # has mean 20, std 1.4

:param N: how many data points in this sample"""                                 # _US
            if isinstance(mean, k1lib.settings.cli.arrayTypes): mean = mean.item() # _US
            if isinstance(std, k1lib.settings.cli.arrayTypes): std = std.item()  # _US
            self.mean = mean; self.std = std; self.N = N                         # _US
        @staticmethod                                                            # _US
        def _sample(mean, std, n=None, _class=0):                                # _US
            t = UValue._unit[_class, random.randint(0, 4)]                       # _US
            if n is not None: t = t[:n]                                          # _US
            return t * std + mean                                                # _US
        def sample(self, n=100, _class=0):                                       # _US
            """Gets a sample :class:`np.ndarray` representative of this
    uncertain value. Example::

        # returns tensor([-5.1095,  3.3117, -2.5759,  ..., -2.5810, -1.8131,  1.8339])
        (k1lib.UValue() * 5).sample()"""                                         # _US
            return UValue._sample(*self, n, _class)                              # _US
        @staticmethod                                                            # _US
        def fromSeries(series, ddof=0):                                          # _US
            """Creates a :class:`UValue` from a bunch of numbers

:param series: can be a list of numbers, numpy array or PyTorch tensor
:param unbiased: if True, Bessel’s correction will be used"""                    # _US
            if hasTorch and isinstance(series, torch.Tensor): series = series.numpy() # _US
            if not isinstance(series, np.ndarray): series = np.array(list(series)) # _US
            series = series * 1.0                                                # _US
            return UValue(series.mean(), np.std(series, ddof=ddof), len(series)) # _US
        @staticmethod                                                            # _US
        def fromBounds(min_, max_):                                              # _US
            """Creates a :class:`UValue` from min and max values.
Example::

    # returns UValue(mean=2.5, std=0.5)
    k1lib.UValue.fromBounds(2, 3)"""                                             # _US
            mid = (min_ + max_)/2                                                # _US
            return k1lib.UValue(mid, abs(max_-mid))                              # _US
        def __iter__(self): yield self.mean; yield self.std                      # _US
        def _niceValue(self, v, _class=0):                                       # _US
            if isinstance(v, UValue): return [UValue._sample(*v, None, _class), UValue._sample(*v, None, _class)] # _US
            return [UValue._sample(v, 0, None, _class), UValue._sample(v, 0, None, _class)] # _US
        def _postProcess(self, c1, c2):                                          # _US
            if c1.hasNan() or c2.hasNan():                                       # _US
                warnings.warn("Calculations has NaN values. They will be replaced with 0, which can affect accuracy of mean and std calculations") # _US
                c1.clearNan(); c2.clearNan()                                     # _US
            c1 = removeOutliers(c1); c2 = removeOutliers(c2);                    # _US
            return UValue(roundOff(c1.mean().item(), c2.mean().item()), roundOff(c1.std().item(), c2.std().item())) # _US
        @property                                                                # _US
        def exact(self):                                                         # _US
            """Whether this UValue is exact or not"""                            # _US
            return self.std == 0                                                 # _US
        @staticmethod                                                            # _US
        def _isValueExact(v):                                                    # _US
            if isinstance(v, UValue): return v.exact                             # _US
            try: len(v); return False                                            # _US
            except: return True                                                  # _US
        @staticmethod                                                            # _US
        def _value(v): # gets mean value                                         # _US
            if isinstance(v, UValue): return v.mean                              # _US
            try: len(v); raise RuntimeError("Can't convert a series into an exact value") # _US
            except: return v                                                     # _US
        def test(self, v):                                                       # _US
            """Returns how many sigma a particular value is."""                  # _US
            return (v-self.mean)/self.std                                        # _US
        def f(self, func):                                                       # _US
            """Covered in :meth:`__init__` docs"""                               # _US
            if self.exact: return UValue(func(self.mean), 0)                     # _US
            f = func; a1, a2 = self._niceValue(self)                             # _US
            try: return self._postProcess(f(a1), f(a2))                          # _US
            except:                                                              # _US
                f = lambda xs: np.array([func(x) for x in xs[:10000]])           # _US
                return self._postProcess(f(a1), f(a2))                           # _US
        def bounds(self):                                                        # _US
            """Returns (mean-std, mean+std)"""                                   # _US
            return self.mean - self.std, self.mean + self.std                    # _US
        def _op2(self, func, a, b):                                              # _US
            if UValue._isValueExact(a) and UValue._isValueExact(b):              # _US
                return UValue(func(UValue._value(a), UValue._value(b)), 0)       # _US
            f = func; a1, a2 = self._niceValue(a, 0); b1, b2 = self._niceValue(b, 1) # _US
            try: return self._postProcess(f(a1, b1), f(a2, b2))                  # _US
            except:                                                              # _US
                f = lambda xs, ys: np.array([func(x, y).item() for x, y in zip(xs[:10000], ys[:10000])]) # _US
                return self._postProcess(f(a1, b1), f(a2, b2))                   # _US
        @staticmethod                                                            # _US
        def combine(*values, samples=1000):                                      # _US
            """Combines multiple UValues into 1.
Example::

    a = k1lib.UValue(5, 1)
    b = k1lib.UValue(7, 1)
    # both returns 6.0 ± 1.4
    k1lib.UValue.combine(a, b)
    [a, b] | k1lib.UValue.combine()

This will sample each UValue by default 1000 times, put them into a
single series and get a UValue from that. Why not just take the
average instead? Because the standard deviation will be less, and
will not actually reflect the action of combining UValues together::

    # returns 6.0 ± 0.7, which is narrower than expected
    (a + b) / 2"""                                                               # _US
            if len(values) == 0: return ~k1lib.cli.aS(UValue.combine)            # _US
            n = len([0 for v in values if v.N == None])                          # _US
            if 0 < n < len(values): raise Exception("Some UValues specified their dataset size (.N variable) and some don't, which makes it impossible to combine these series together. Either specify the dataset size (change .N value) to specific numbers for all of them, or set them all to None, in which case each series importance will be set to the same level") # _US
            means = [v.mean for v in values]; stds = [v.std for v in values]     # _US
            Ns = [(v.N or 1) for v in values]; N = sum(Ns)                       # _US
            mean = sum(m*n for m,n in zip(means, Ns))/N                          # _US
            std = (sum(n*(s**2 + (m-mean)**2) for m,s,n in zip(means, stds, Ns))/N)**0.5 # _US
            return UValue(mean, std, N)                                          # _US
        def __add__(self, v):                                                    # _US
            m1, s1 = _US(self); m2, s2 = _US(v)                                  # _US
            return UValue(m1+m2, math.sqrt(s1**2 + s2**2))                       # _US
            return self._op2(lambda a, b: a+b, v, self) # representative of how this would work stochastically # _US
        def __radd__(self, v):                                                   # _US
            m1, s1 = _US(self); m2, s2 = _US(v)                                  # _US
            return UValue(m1+m2, math.sqrt(s1**2 + s2**2))                       # _US
        def __sub__(self, v):                                                    # _US
            m1, s1 = _US(self); m2, s2 = _US(v)                                  # _US
            return UValue(m1-m2, math.sqrt(s1**2 + s2**2))                       # _US
        def __rsub__(self, v):                                                   # _US
            m1, s1 = _US(self); m2, s2 = _US(v)                                  # _US
            return UValue(m2-m1, math.sqrt(s1**2 + s2**2))                       # _US
        def __mul__(self, v):                                                    # _US
            m1, s1 = _US(self); m2, s2 = _US(v)                                  # _US
            return UValue(m1*m2, math.sqrt(m2**2*s1**2 + m1**2*s2**2))           # _US
        def __rmul__(self, v):                                                   # _US
            m1, s1 = _US(self); m2, s2 = _US(v)                                  # _US
            return UValue(m1*m2, math.sqrt(m2**2*s1**2 + m1**2*s2**2))           # _US
        def __truediv__(self, v):                                                # _US
            m1, s1 = _US(self); m2, s2 = _US(v)                                  # _US
            return UValue(m1/m2, math.sqrt(1/m2**2*s1**2 + m1**2/m2**4*s2**2))   # _US
        def __rtruediv__(self, v):                                               # _US
            m1, s1 = _US(v); m2, s2 = _US(self)                                  # _US
            return UValue(m1/m2, math.sqrt(1/m2**2*s1**2 + m1**2/m2**4*s2**2))   # _US
        def __pow__(self, v):                                                    # _US
            m1, s1 = _US(self); m2, s2 = _US(v); m = m1**m2                      # _US
            return UValue(m, math.sqrt((m2*m/m1)**2*s1**2 + (math.log(m1)*m)**2*s2**2)) # _US
        def __rpow__(self, v):                                                   # _US
            m1, s1 = _US(v); m2, s2 = _US(self); m = m1**m2                      # _US
            return UValue(m, math.sqrt((m2*m/m1)**2*s1**2 + (math.log(m1)*m)**2*s2**2)) # _US
        def __abs__(self): return self.f(lambda a: abs(a)) # can't convert to pure math that makes sense # _US
        def __neg__(self): return 0 - self                                       # _US
        def __lt__(self, v): return self.mean < v.mean if isinstance(v, UValue) else self.mean < v # _US
        def __eq__(self, v): return self.mean == v.mean and self.std == v.std and self.N == v.N if isinstance(v, UValue) else self.mean == v # _US
        def __float__(self): return self.mean                                    # _US
        def __int__(self): return int(self.mean)                                 # _US
        def __str__(self): mean, std = niceUS(self.mean, self.std); return f"{mean} ± {std}" # _US
        def __repr__(self):                                                      # _US
            mean, std = niceUS(self.mean, self.std)                              # _US
            return f"UValue(mean={toPrecision(self.mean, 5)}, std={toPrecision(self.std, 5)}) -> {mean} ± {std}" # _US
        def plot(self, name=None):                                               # _US
            """Quickly plots a histogram of the distribution.
Possible to plot multiple histograms in 1 plot."""                               # _US
            plt.hist(self.sample(None).numpy(), bins=100, alpha=0.7, label=name) # _US
            if name != None: plt.legend()                                        # _US
if hasTorch:                                                                     # _US
    @k1lib.patch(torch.Tensor)                                                   # _US
    def clearNan(self, value:float=0.0) -> torch.Tensor:                         # _US
        """Sets all nan values to a specified value.
Example::

    a = torch.randn(3, 3) * float("nan")
    a.clearNan() # now full of zeros"""                                          # _US
        self[self != self] = value                                               # _US
        return self                                                              # _US
class ConstantPad:                                                               # ConstantPad
    def __init__(self, left=False):                                              # ConstantPad
        """Adds constant amount of padding to strings.
Example::

    p = k1.ConstantPad()
    p("123")    # returns "123"
    p("23")     # returns " 23"
    "12345" | p # returns "12345", can pipe it in too, but is not strictly a cli tool
    p("123")    # returns "  123"

Basically, this is useful in situations when you're printing a table or status bar and
needs relatively constant width but you don't know what's the desired width at the start.

As you normally use a bunch of these in groups, there's a convenience function for
that too::

    p1, p2 = k1.ConstantPad.multi(2)

:param left: whether to align left or not"""                                     # ConstantPad
        self.left = left; self.length = 0                                        # ConstantPad
    def __call__(self, s):                                                       # ConstantPad
        self.length = max(self.length, len(s))                                   # ConstantPad
        return s.ljust(self.length) if self.left else s.rjust(self.length)       # ConstantPad
    def __ror__(self, s): return self.__call__(s)                                # ConstantPad
    @staticmethod                                                                # ConstantPad
    def multi(n, *args, **kwargs): return [ConstantPad(*args, **kwargs) for i in range(n)] # ConstantPad



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
from .callbacks import Callback, Callbacks, Cbs
import k1lib, time, math, logging, numpy as np
from functools import partial; plt = k1lib.dep("matplotlib.pyplot")
try: import torch; from torch import nn; hasTorch = True
except: hasTorch = False
__all__ = ["Profiler"]
if hasTorch:
    import k1lib.callbacks.profilers as ps
    ComputationProfiler = ps.computation.ComputationProfiler
    IOProfiler = ps.io.IOProfiler
    MemoryProfiler = ps.memory.MemoryProfiler
    TimeProfiler = ps.time.TimeProfiler
else:
    class ComputationProfiler: pass
    class IOProfiler: pass
    class MemoryProfiler: pass
    class TimeProfiler: pass
@k1lib.patch(Cbs)
class Profiler(Callback):                                                        # Profiler
    """Profiles memory, time, and computational complexity of the network. See over
:mod:`k1lib.callbacks.profilers` for more details on each of these profilers"""  # Profiler
    def __init__(self):                                                          # Profiler
        super().__init__(); self.clear(); self.dependsOn=["Recorder"]            # Profiler
    def clear(self):                                                             # Profiler
        """Clears every child profilers"""                                       # Profiler
        self._mpCache=None; self._tpCache=None                                   # Profiler
        self._cpCache=None; self._ioCache=None                                   # Profiler
    def _memory(self): # do this to quickly debug, cause if not, Callback will just raise AttributeError on .memory # Profiler
        if self._mpCache != None: return self._mpCache                           # Profiler
        with self.cbs.context():                                                 # Profiler
            mp = MemoryProfiler(); self.cbs.add(mp)                              # Profiler
            mp._run(); self._mpCache = mp; return mp                             # Profiler
    @property                                                                    # Profiler
    def memory(self) -> MemoryProfiler:                                          # Profiler
        """Gets :class:`~k1lib.callbacks.profilers.memory.MemoryProfiler`"""     # Profiler
        return self._memory()                                                    # Profiler
    def _computation(self):                                                      # Profiler
        if self._cpCache != None: return self._cpCache                           # Profiler
        with self.cbs.context():                                                 # Profiler
            cp = ComputationProfiler(self); self.cbs.add(cp)                     # Profiler
            cp._run(); self._cpCache = cp; return cp                             # Profiler
    @property                                                                    # Profiler
    def computation(self) -> ComputationProfiler:                                # Profiler
        """Gets :class:`~k1lib.callbacks.profilers.computation.ComputationProfiler`""" # Profiler
        return self._computation()                                               # Profiler
    def _time(self):                                                             # Profiler
        if self._tpCache != None: return self._tpCache                           # Profiler
        with self.cbs.context():                                                 # Profiler
            tp = TimeProfiler(); self.cbs.add(tp)                                # Profiler
            tp._run(); self._tpCache = tp; return tp                             # Profiler
    @property                                                                    # Profiler
    def time(self) -> TimeProfiler:                                              # Profiler
        """Gets :class:`~k1lib.callbacks.profilers.time.TimeProfiler`"""         # Profiler
        return self._time()                                                      # Profiler
    def _io(self):                                                               # Profiler
        if self._ioCache != None: return self._ioCache                           # Profiler
        with self.cbs.context():                                                 # Profiler
            io = IOProfiler(); self.cbs.add(io)                                  # Profiler
            io._run(); self._ioCache = io; return io                             # Profiler
    @property                                                                    # Profiler
    def io(self) -> IOProfiler:                                                  # Profiler
        """Gets :class:`~k1lib.callbacks.profilers.io.IOProfiler`"""             # Profiler
        return self._io()                                                        # Profiler
    def __repr__(self):                                                          # Profiler
        return f"""{self._reprHead}, can...
- p.memory: to profile module memory requirements
- p.time: to profile module execution times
- p.computation: to estimate module computation
- p.io: to get input and output shapes of
{self._reprCan}"""                                                               # Profiler



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""This module is for all things related to atoms, molecules and their simulations"""
import k1lib, math, k1lib.cli as cli
from typing import Dict, List, Union
plt = k1lib.dep("matplotlib.pyplot")
from collections import deque
from .atom import Atom, _depthAuto
try: import torch; hasTorch = True
except: torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["System"]
def distV(x:torch.Tensor) -> torch.Tensor:                                       # distV
    """Distance vectors of points.

:param x: location Tensor of shape (n, 3)
:returns: vector Tensor of shape (n, n, 3)"""                                    # distV
    n = x.shape[0]                                                               # distV
    return x.view(1, n, 3) - x.view(n, 1, 3)                                     # distV
@k1lib.patch(Atom)                                                               # distV
def _s_bondLength(self, idx_i:Dict[str, int], bondLengths:torch.Tensor):         # _s_bondLength
    """Calculates bond length for all bonds in this atom and stores in
``bondLengths``."""                                                              # _s_bondLength
    bonds = list(set(self.bonds)) + self.eClouds                                 # _s_bondLength
    for atom in bonds:                                                           # _s_bondLength
        nBonds = self.nBonds(atom)                                               # _s_bondLength
        bL = self.radius[nBonds-1] + atom.radius[nBonds-1]                       # _s_bondLength
        bondLengths[idx_i[self.idx], idx_i[atom.idx]] = bL                       # _s_bondLength
        bondLengths[idx_i[atom.idx], idx_i[self.idx]] = bL                       # _s_bondLength
class System:                                                                    # System
    def __init__(self, mapping:Dict[str, Atom]):                                 # System
        """Creates a new system that contains a molecule so that it can be
simulated. Not intended to be used by the end user. Use :meth:`Atom.system`
instead.

:param mapping: maps from atom index to :class:`Atom` object"""                  # System
        self.mapping = mapping; self.atoms = list(mapping.values())              # System
        self.i_idx = list(mapping.keys()); self.n = len(self.i_idx)              # System
        self.idx_i = {idx:i for i, idx in enumerate(self.i_idx)}                 # System
        self._bondLengths = torch.zeros(self.n, self.n)                          # System
        for atom in self.atoms: atom._s_bondLength(self.idx_i, self._bondLengths) # System
        self._graphDistances = self._calcGraphDistances()                        # System
        self._bondMask = (self._bondLengths > 0) + 0                             # System
        self._x = 50*torch.randn(self.n, 3)                                      # System
        self._v = torch.zeros(self.n, 3)                                         # System
        self._a = torch.zeros(self.n, 3)                                         # System
        self.iMask = 1-torch.eye(self.n) # identity mask                         # System
    def __getitem__(self, idx:Union[str, int, Atom]) -> Union[Atom, Atom, int]:  # System
        """Look up Atom that has idx (str/int). If Atom is passed in, look
for the index of it."""                                                          # System
        if isinstance(idx, str): return self.mapping[idx]                        # System
        if isinstance(idx, int): return self.mapping[self.i_idx[idx]]            # System
        if isinstance(idx, Atom): return self.idx_i[idx.idx]                     # System
        raise RuntimeError(f"Don't understand input of type {type(idx)}")        # System
    def x(self, a:Atom) -> torch.Tensor:                                         # System
        """Get current location vector of the specified :class:`Atom`."""        # System
        return self._x[self.idx_i[a.idx]]                                        # System
    def v(self, a:Atom):                                                         # System
        """Get current velocity vector of the specified :class:`Atom`."""        # System
        return self._v[self.idx_i[a.idx]]                                        # System
@k1lib.patch(Atom)                                                               # System
def _s_setGD(self:Atom, s:System, atom:Atom, gd:torch.Tensor, v:int, todos:deque, autoCount): # _s_setGD
    """
:param a: other (constant) atom that I'm trying to get distance to
:param gd: graph distances
:param v: value (aka distance)
:param todos: queue of atoms to be processed
:param autoCount: to measure perf"""                                             # _s_setGD
    if v > 4: return                                                             # _s_setGD
    x = self._s_gi; y = atom._s_gi # int indexes in system                       # _s_setGD
    if v < gd[x, y]:                                                             # _s_setGD
        autoCount(); gd[y, x] = gd[x, y] = v                                     # _s_setGD
        for a in self.uniqueBonds + self.eClouds:                                # _s_setGD
            todos.append((a, self, atom)) # tuple (atom to be updated, (what happens here)) # _s_setGD
        for a in atom.uniqueBonds + atom.eClouds:                                # _s_setGD
            todos.append((a, atom, self))                                        # _s_setGD
@k1lib.patch(System)                                                             # _s_setGD
def _calcGraphDistances(self):                                                   # _calcGraphDistances
    gd = torch.ones(self.n, self.n) * float("inf")                               # _calcGraphDistances
    todos = deque(); self._gDAutoCount = k1lib.AutoIncrement()                   # _calcGraphDistances
    for i, a in enumerate(self.atoms): a._s_gi = i # cached i value              # _calcGraphDistances
    for a in self.atoms: a._s_setGD(self, a, gd, 0, todos, self._gDAutoCount)    # _calcGraphDistances
    while len(todos) > 0:                                                        # _calcGraphDistances
        mainA, _as, _ad = todos.popleft() # "source" and "destination"           # _calcGraphDistances
        mainA._s_setGD(self, _ad, gd, gd[_as._s_gi, _ad._s_gi] + 1, todos, self._gDAutoCount) # _calcGraphDistances
    return gd                                                                    # _calcGraphDistances
@k1lib.patch(System)                                                             # _calcGraphDistances
def _calcForces(self):                                                           # _calcForces
    n = self.n; dV = distV(self._x) # (n, n, 3)                                  # _calcForces
    l = torch.sqrt((dV**2).sum(dim=2)) + 0.001                                   # _calcForces
    dV = dV / l.view(n, n, 1) # normalized direction                             # _calcForces
    aBond = 10 * dV * torch.tanh((l-self._bondLengths)*self._bondMask/100).view(n, n, 1) # _calcForces
    vBond = aBond.sum(dim=1)                                                     # _calcForces
                                                                                 # _calcForces
    aCou = dV * ((100/l)**4).view(n, n, 1) # coulomb forces on every atom        # _calcForces
    vCou = .03 * aCou.clearNan().sum(dim=1)                                      # _calcForces
                                                                                 # _calcForces
    aCouC = dV * ((100/l)**2).view(n, n, 1) # close coulomb forces               # _calcForces
    couCMask = ((self._graphDistances < 4) + 0).view(n, n, 1)                    # _calcForces
    vCouC = .3 * (aCouC * couCMask).clearNan().sum(dim=1)                        # _calcForces
                                                                                 # _calcForces
    self._a = 10 * (vBond - vCou - vCouC)                                        # _calcForces
    return aBond, aCou, l, dV, vBond, vCou # return stuff for debugging purposes # _calcForces
@k1lib.patch(System)                                                             # _calcForces
def _update(self, dt:float=1.0):                                                 # _update
    self._v += dt * self._a; self._a.zero_()                                     # _update
    self._v = torch.clamp(self._v * 0.98, -10, 10) # friction                    # _update
    self._x += dt * self._v; self._x *= 0.999 # universe wants to quish things down # _update
@k1lib.patch(System)                                                             # _update
def _changeDevice(self, device:str="cpu"):                                       # _changeDevice
    self._bondLengths = self._bondLengths.to(device)                             # _changeDevice
    self._graphDistances = self._graphDistances.to(device)                       # _changeDevice
    self._bondMask = self._bondMask.to(device)                                   # _changeDevice
    self._x = self._x.to(device)                                                 # _changeDevice
    self._v = self._v.to(device)                                                 # _changeDevice
    self._a = self._a.to(device)                                                 # _changeDevice
    self.iMask = self.iMask.to(device)                                           # _changeDevice
@k1lib.patch(System)                                                             # _changeDevice
def simulate(self, t:int=300, dt:float=1.0, recordXs:bool=True, cuda:bool=False) -> List[torch.Tensor]: # simulate
    """Simulate system for ``t`` steps.

:param t: simulation total steps. 100 to view dynamics closely, 1000 to make
    sure it converges, default 300 is sweet spot
:param dt: simulation time between steps
:param recordXs: whether to record locations while the simulation happens. Put
    False to save memory/performance.
:param cuda: if True, do the simulation on the graphics card
:return: if ``recordXs`` is True, returns max 100 location Tensors. The Tensors
    will have shape of (n, 3), where n is the number of atoms and electron clouds
    your molecule has."""                                                        # simulate
    self._changeDevice("cuda" if cuda else "cpu")                                # simulate
    every = k1lib.Every((t // 100) or 1); xs = []                                # simulate
    for i in range(t):                                                           # simulate
        self._calcForces(); self._update(dt)                                     # simulate
        if recordXs and every(): xs.append(self._x.clone())                      # simulate
    return xs                                                                    # simulate
@k1lib.patch(Atom)                                                               # simulate
def _system(self, mapping:Dict[str, Atom], gDepth:int):                          # _system
    if self.gDepth >= gDepth: return                                             # _system
    self.gDepth = gDepth; mapping[self.idx] = self                               # _system
    for atom in self.bonds: atom._system(mapping, gDepth)                        # _system
    for eCloud in self.eClouds: eCloud._system(mapping, gDepth)                  # _system
@k1lib.patch(Atom)                                                               # _system
def system(self) -> System:                                                      # system
    """Creates a :class:`System` of the molecule this :class:`Atom` is attached
to."""                                                                           # system
    mapping = dict(); self._system(mapping, _depthAuto()); return System(mapping) # system
@k1lib.patch(Atom)                                                               # system
def _s_plot(self, ax, x:torch.Tensor, idx_i:Dict[str, int], H:bool):             # _s_plot
    if not H and self.name == "H": return                                        # _s_plot
    nH = len(self.HBonds); HName = "" if nH == 0 else ("H" if nH == 1 else f"H{nH}") # _s_plot
    s = x[idx_i[self.idx]]                                                       # _s_plot
    ax.text(s[0], s[1], s[2], f"{self.name + ('' if H else HName)}", ha="center", va="center") # _s_plot
    for atom in (self.uniqueBonds if H else self.uniqueNonHBonds):               # _s_plot
        a = x[idx_i[atom.idx]]                                                   # _s_plot
        ax.plot([s[0], a[0]], [s[1], a[1]], [s[2], a[2]])                        # _s_plot
@k1lib.patch(System)                                                             # _s_plot
def plot(self, x:torch.Tensor=None, ax:"matplotlib.axes.Axes"=None, H:bool=True): # plot
    """Plots the molecule.
Example::

    %matplotlib widget
    s = mo.CH4(mo.H2O).system()
    s.simulate(); s.plot()

The first line is so that you can rotate the molecule around in 3d interactively.
The 3rd line includes a simulation step, to get the molecule into roughly the
right position.

:param x: location Tensor of shape (n, 3). If none provided, will use current locations
:param ax: Axes object, in case you want to plot this on an existing plot
:param H: whether to include Hydrogens as separate atoms or bunched together to
    another main atom."""                                                        # plot
    if x is None: x = self._x                                                    # plot
    if ax is None: ax = plt.figure(dpi=150).add_subplot(projection="3d")         # plot
    ax.clear(); com = x.mean(dim=0) # center of mass                             # plot
    box = math.sqrt(((x - com.view(1, -1))**2).sum(dim=1).max())                 # plot
    ax.set_xlim(com[0]-box, com[0]+box)                                          # plot
    ax.set_ylim(com[1]-box, com[1]+box)                                          # plot
    ax.set_zlim(com[2]-box, com[2]+box)                                          # plot
    for atom in self.atoms:                                                      # plot
        if atom.name != "_e": atom._s_plot(ax, x, self.idx_i, H)                 # plot
    return ax                                                                    # plot
@k1lib.patch(System)                                                             # plot
def animate(self, xs:List[torch.Tensor], rotateSpeed=0.5, H:bool=True):          # animate
    """Animates the molecule. This requires location information from
:meth:`simulate`. Example::

    s = mo.CH4(mo.H2O).system()
    s.animate(s.simulate())

:param H: whether to include Hydrogens as separate atoms or bunched together to
    another main atom."""                                                        # animate
    fig = plt.figure(); ax = fig.add_subplot(projection="3d"); plt.close(fig)    # animate
    def update(frame):                                                           # animate
        self.plot(xs[frame], ax, H); ax.view_init(frame*rotateSpeed, frame*2*rotateSpeed) # animate
    a = k1lib.viz.FAnim(fig, update, len(xs)); plt.close(fig); return a          # animate
